import tensorflow as tf

from seq2annotation.algorithms.ngram_model import NgramModel


class BilstmCrfNgramModel(NgramModel):
    def bilstm_layer(self, embeddings, nwords):
        t = tf.transpose(embeddings, perm=[1, 0, 2])
        lstm_cell_fw = tf.contrib.rnn.LSTMBlockFusedCell(self.params['lstm_size'])
        lstm_cell_bw = tf.contrib.rnn.LSTMBlockFusedCell(self.params['lstm_size'])
        lstm_cell_bw = tf.contrib.rnn.TimeReversedFusedRNN(lstm_cell_bw)
        output_fw, _ = lstm_cell_fw(t, dtype=tf.float32,
                                    sequence_length=nwords)
        output_bw, _ = lstm_cell_bw(t, dtype=tf.float32,
                                    sequence_length=nwords)

        # lstm_fw_cell = tf.nn.rnn_cell.BasicLSTMCell(self.params['lstm_size'])
        # lstm_bw_cell = tf.nn.rnn_cell.BasicLSTMCell(self.params['lstm_size'])
        # outputs, output_fw, output_bw = tf.nn.bidirectional_dynamic_rnn(
        #     lstm_fw_cell,
        #     lstm_bw_cell, t,
        #     sequence_length=nwords,
        #     dtype=tf.float32
        # )

        output = tf.concat([output_fw, output_bw], axis=-1)
        # transpose it back
        output = tf.transpose(output, perm=[1, 0, 2])

        return output

    def call(self, embeddings, nwords):
        data = self.bilstm_layer(embeddings, nwords)
        return data



import tensorflow as tf
from seq2annotation.algorithms.model import Model


class StackedBilstmCrfModel(Model):
    @classmethod
    def default_params(cls):
        default_params = {
            'stacked_layers': 2
        }

        return default_params

    def bilstm_layer(self, embeddings, nwords):
        t = tf.transpose(embeddings, perm=[1, 0, 2])
        lstm_cell_fw = tf.contrib.rnn.LSTMBlockFusedCell(self.params['lstm_size'])
        lstm_cell_bw = tf.contrib.rnn.LSTMBlockFusedCell(self.params['lstm_size'])
        lstm_cell_bw = tf.contrib.rnn.TimeReversedFusedRNN(lstm_cell_bw)
        output_fw, _ = lstm_cell_fw(t, dtype=tf.float32,
                                    sequence_length=nwords)
        output_bw, _ = lstm_cell_bw(t, dtype=tf.float32,
                                    sequence_length=nwords)
        output = tf.concat([output_fw, output_bw], axis=-1)
        # transpose it back
        output = tf.transpose(output, perm=[1, 0, 2])

        return output

    def call(self, embeddings, nwords):
        inner_layer_data = self.bilstm_layer(embeddings, nwords)
        for i in range(1, self.params['stacked_layers']):
            inner_layer_data = self.bilstm_layer(inner_layer_data, nwords)

        return inner_layer_data



from pathlib import Path

import tensorflow as tf
import numpy as np
from seq2annotation.metrics import precision, recall, f1
from tokenizer_tools.tagset.NER.BILUO import BILUOEncoderDecoder


class LookupModel(object):
    @classmethod
    def default_params(cls):
        return {}

    @classmethod
    def model_fn(cls, features, labels, mode, params):
        instance = cls(features, labels, mode, params)
        return instance()

    def __init__(self, features, labels, mode, params):
        self.features = features
        self.labels = labels
        self.mode = mode
        self.params = params

    def input_layer(self):
        data = np.loadtxt(self.params['words'], dtype=np.unicode, encoding=None)
        mapping_strings = tf.Variable(data.reshape((-1,)))
        vocab_words = tf.contrib.lookup.index_table_from_tensor(
            mapping_strings, num_oov_buckets=self.params['num_oov_buckets'])

        # Word Embeddings
        words = tf.identity(self.features['words'], name='input_words')
        word_ids = vocab_words.lookup(words)


        #
        # raw_nwords = tf.identity(features['words_len'], name='input_words_len')
        # nwords = tf.feature_column.input_layer({'words_len': raw_nwords}, params['words_len_feature_columns'])
        # nwords = tf.reshape(nwords, [-1])
        # nwords = tf.to_int32(nwords)

        # words = features['words']
        # words = tf.convert_to_tensor(words)
        #
        # nwords = features['words_len']
        # nwords = tf.convert_to_tensor(nwords)

        nwords = tf.identity(self.features['words_len'], name='input_words_len')

        # get tag info
        with Path(self.params['tags']).open() as f:
            indices = [idx for idx, tag in enumerate(f) if tag.strip() != 'O']
            num_tags = len(indices) + 1

        # # true tags to ids
        # if self.mode == tf.estimator.ModeKeys.PREDICT:
        #     true_tag_ids = 0
        # else:
        #     true_tag_ids = self.tag2id(self.labels)

        return indices, num_tags, word_ids, nwords

    def input_lookup_layer(self):
        # data = np.loadtxt(self.params['lookup'], dtype=np.unicode, encoding=None)
        # mapping_strings = tf.Variable(data.reshape((-1,)))
        # vocab_words = tf.contrib.lookup.index_table_from_tensor(
        #     mapping_strings, num_oov_buckets=self.params['num_oov_buckets'])
        #
        # # Word Embeddings
        # words = tf.identity(self.features['lookup'], name='lookup')
        # word_ids = vocab_words.lookup(words)
        #
        # return word_ids

        raw_feature = self.features['lookup']

        raw_feature_shape = tf.shape(raw_feature)
        flat_feature = tf.reshape(raw_feature, [-1])

        feature_column = tf.feature_column.indicator_column(
            tf.feature_column.categorical_column_with_vocabulary_list(
                key='lookup',
                vocabulary_list=[True, False])
        )

        encoded_feature = tf.feature_column.input_layer({'lookup': flat_feature}, [feature_column])

        raw_encoded_feature = tf.shape(encoded_feature)

        feature_shape = [raw_feature_shape[0], raw_feature_shape[1], 5]

        feature = tf.reshape(encoded_feature, feature_shape)

        return feature

    def embedding_layer(self, word_ids):
        # load pre-trained data from file
        # glove = np.load(params['glove'])['embeddings']  # np.array

        # training the embedding during training
        glove = np.zeros((self.params['embedding']['vocabulary_size'], self.params['dim']), dtype=np.float32)

        # Add OOV word embedding
        embedding_array = np.vstack([glove, [[0.] * self.params['dim']]])

        embedding_variable = tf.Variable(embedding_array, dtype=tf.float32, trainable=True)
        embeddings = tf.nn.embedding_lookup(embedding_variable, word_ids)

        return embeddings

    def dropout_layer(self, data):
        training = (self.mode == tf.estimator.ModeKeys.TRAIN)
        output = tf.layers.dropout(data, rate=self.params['dropout'], training=training)

        return output

    def dense_layer(self, data, num_tags):
        logits = tf.layers.dense(data, num_tags)

        return logits

    def load_tag_data(self):
        data = np.loadtxt(self.params['tags'], dtype=np.unicode, encoding=None)
        mapping_strings = tf.Variable(data.reshape((-1,)))

        return mapping_strings

    def tag2id(self, labels):
        mapping_strings = self.load_tag_data()
        vocab_tags = tf.contrib.lookup.index_table_from_tensor(mapping_strings)

        tags = vocab_tags.lookup(labels)

        return tags

    def id2tag(self, pred_ids):
        mapping_strings = self.load_tag_data()
        reverse_vocab_tags = tf.contrib.lookup.index_to_string_table_from_tensor(
            mapping_strings)

        pred_strings = reverse_vocab_tags.lookup(tf.to_int64(pred_ids))

        return pred_strings

    def loss_layer(self, preds, ground_true, nwords, crf_params):
        with tf.name_scope("CRF_log_likelihood"):
            log_likelihood, _ = tf.contrib.crf.crf_log_likelihood(
                preds, ground_true, nwords, crf_params)

        loss = tf.reduce_mean(-log_likelihood)

        return loss

    def crf_decode_layer(self, logits, crf_params, nwords):
        with tf.name_scope("CRF_decode"):
            pred_ids, _ = tf.contrib.crf.crf_decode(logits, crf_params,
                                                        nwords)

        return pred_ids

    def compute_metrics(self, tags, pred_ids, num_tags, indices, nwords):
        weights = tf.sequence_mask(nwords)
        metrics = {
            'acc': tf.metrics.accuracy(tags, pred_ids, weights),
            'precision': precision(tags, pred_ids, num_tags, indices,
                                   weights),
            'recall': recall(tags, pred_ids, num_tags, indices, weights),
            'f1': f1(tags, pred_ids, num_tags, indices, weights),
        }

        for metric_name, op in metrics.items():
            tf.summary.scalar(metric_name, op[1])

        return metrics

    def call(self, embeddings, nwords):
        raise NotImplementedError

    def __call__(self):
        indices, num_tags, word_ids, nwords = self.input_layer()
        lookup_feature = self.input_lookup_layer()

        embeddings = self.embedding_layer(word_ids)

        feature_data = tf.concat((embeddings, lookup_feature), axis=2)

        data = self.call(feature_data, nwords)

        data = self.dropout_layer(data)
        logits = self.dense_layer(data, num_tags)

        crf_params = tf.get_variable("crf", [num_tags, num_tags],
                                     dtype=tf.float32)

        pred_ids = self.crf_decode_layer(logits, crf_params, nwords)

        pred_strings = self.id2tag(pred_ids)
        predictions = {
            'pred_ids': pred_ids,
            'tags': pred_strings
        }

        if self.mode == tf.estimator.ModeKeys.PREDICT:
            return tf.estimator.EstimatorSpec(self.mode,
                                              predictions=predictions)
        else:
            true_tag_ids = self.tag2id(self.labels)

            loss = self.loss_layer(logits, true_tag_ids, nwords, crf_params)

            metrics = self.compute_metrics(true_tag_ids, pred_ids,
                                           num_tags, indices, nwords)

            if self.mode == tf.estimator.ModeKeys.EVAL:
                return tf.estimator.EstimatorSpec(
                    self.mode, loss=loss, eval_metric_ops=metrics)

            elif self.mode == tf.estimator.ModeKeys.TRAIN:
                train_op = tf.train.AdamOptimizer().minimize(
                    loss, global_step=tf.train.get_or_create_global_step())
                return tf.estimator.EstimatorSpec(
                    self.mode, loss=loss, train_op=train_op)



from typing import List

import keras

from tokenizer_tools.tagset.NER.BILUO import BILUOSequenceEncoderDecoder
from tokenizer_tools.tagset.offset.sequence import Sequence
from tensorflow.contrib import predictor

from tokenizer_tools.tagset.exceptions import TagSetDecodeError

decoder = BILUOSequenceEncoderDecoder()


class Inference(object):
    def __init__(self, model_path):
        # load model
        self.model_dir = model_path
        self.predict_fn = predictor.from_saved_model(model_path)

    def infer(self, input_text: str):
        infer_result = self._infer(input_text)
        return infer_result[0]

    def batch_infer(self, input_text: List[str]):
        return self._infer(input_text)

    def _infer(self, input_text):
        if isinstance(input_text, str):
            input_list = [input_text]
        else:
            input_list = input_text

        raw_sequences = [[i for i in text] for text in input_list]

        sentence = keras.preprocessing.sequence.pad_sequences(
            raw_sequences, dtype='object',
            padding='post', truncating='post', value=['<pad>']
        ).tolist()

        # TODO: batch infer will cause padding, which will maybe cause decoder to offset bug.
        # TODO: feature translate should out of this main program for better compatible with keras and estimator model
        input_feature = {
            'words': [[i for i in text] for text in sentence],
            'words_len': [len(text) for text in raw_sequences],
        }

        # print(input_feature)

        predictions = self.predict_fn(input_feature)
        tags_list = predictions['tags']

        infer_result = []
        for raw_input_text, raw_text, normalized_text, tags in zip(input_list, raw_sequences, sentence, tags_list):
            # decode Unicode
            tags_seq = [i.decode() for i in tags]

            # print(tags_seq)

            # BILUO to offset
            failed = False
            try:
                seq = decoder.to_offset(tags_seq, raw_text)
            except TagSetDecodeError as e:
                print(e)

                # invalid tag sequence will raise exception
                # so return a empty result
                seq = Sequence(input_text)
                failed = True

            infer_result.append((raw_input_text, seq, tags_seq, failed))

        return infer_result



from typing import List, Union, Any, Callable

import numpy as np

from tokenizer_tools.tagset.NER.BILUO import BILUOSequenceEncoderDecoder
from tokenizer_tools.tagset.offset.sequence import Sequence

from tokenizer_tools.tagset.exceptions import TagSetDecodeError

decoder = BILUOSequenceEncoderDecoder()


class InferenceBase(object):
    def __init__(self, model_path, *args, **kwargs):
        self.model_path = model_path

        self.predict_fn = None
        self.load_prediction_fn()

    def load_prediction_fn(self):
        self.predict_fn = self.instance_predict_fn()

    def instance_predict_fn(self) -> Callable:
        raise NotImplementedError

    def infer(self, input_text: str):
        input_text_list = [input_text]
        infer_result = self.do_infer(input_text_list)
        return infer_result[0]

    def batch_infer(self, input_text: List[str]):
        return self.do_infer(input_text)

    def preprocess(
        self, msg_list: List[List[str]]
    ) -> Union[np.ndarray, List[List[str]]]:
        raise NotImplementedError

    def encode_input_feature(self, msg_list: Union[np.ndarray, List[List[str]]]) -> Any:
        raise NotImplementedError

    def decode_output_feature(self, response: Any) -> Any:
        raise NotImplementedError

    def postprocess(self, input_: Any) -> List[List[str]]:
        raise NotImplementedError

    def decode_ner_tag_sequence(
        self,
        ner_tag_sequence: List[List[str]],
        std_msg_list: List[List[str]],
        raw_msg_list: List[str],
    ):
        infer_result = []
        for raw_msg, std_msg, ner_tags in zip(
            raw_msg_list, std_msg_list, ner_tag_sequence
        ):
            failed = False
            try:
                # BILUO to offset
                seq = decoder.to_offset(ner_tags, std_msg)
            except TagSetDecodeError as e:
                print(e)

                # invalid tag sequence will raise exception
                # so return a empty result
                seq = Sequence(std_msg)
                failed = True

            infer_result.append((raw_msg, seq, ner_tags, failed))

        return infer_result

    def do_infer(self, input_list: List[str]):
        std_input_list = [[i for i in text] for text in input_list]

        # do work, such like str2id and padding et al.
        preprocessed_msg = self.preprocess(std_input_list)

        # compose data structure and adding feature engineering data
        encoded_input_feature = self.encode_input_feature(preprocessed_msg)

        raw_predictions = self.predict_fn(encoded_input_feature)

        # extract predictions value from raw_predictions
        predictions = self.decode_output_feature(raw_predictions)

        # do work, such like id2str and so on
        ner_tag_sequence = self.postprocess(predictions)

        # decode ner tag and compose final result for return
        result = self.decode_ner_tag_sequence(
            ner_tag_sequence, std_input_list, input_list
        )

        return result



import os
import time

import paddle
import paddle.fluid as fluid


def train_model(train_inpf, eval_inpf, config, *args, **kwargs):
    main_program = fluid.Program()
    startup_program = fluid.Program()
    with fluid.program_guard(main_program=main_program, startup_program=startup_program):
        words = fluid.layers.data(name='words', shape=[1], dtype='int64',
                                  lod_level=1)
        tags = fluid.layers.data(name='tags', shape=[1], dtype='int64',
                                 lod_level=1)

        embed_char = fluid.layers.embedding(
            input=words,
            size=[config['embedding_vocabulary_size'], config['embedding_dim']],
            dtype='float32')

        place = fluid.CPUPlace()

        train_reader = paddle.batch(
            paddle.reader.shuffle(train_inpf, buf_size=500),
            batch_size=config['batch_size'])

        forward_hidden_state, _ = fluid.layers.dynamic_lstm(
            input=embed_char,
            size=config['embedding_dim'],  # TODO: paddle don't support hidden_size != embedding size
            candidate_activation='relu',
            gate_activation='sigmoid',
            cell_activation='sigmoid')

        backward_hidden_state, _ = fluid.layers.dynamic_lstm(
            input=embed_char,
            size=config['embedding_dim'],
            candidate_activation='relu',
            gate_activation='sigmoid',
            cell_activation='sigmoid',
            is_reverse=True)

        hidden_state = fluid.layers.concat(
            [forward_hidden_state, backward_hidden_state], 1)

        feature = fluid.layers.dropout(hidden_state, 0.1)

        score = fluid.layers.fc(feature, len(config['tags_data']))

        crf_cost = fluid.layers.linear_chain_crf(
            input=score,
            label=tags,
            param_attr=fluid.ParamAttr(name='crfw')
        )

        avg_cost = fluid.layers.mean(crf_cost)

        crf_decode = fluid.layers.crf_decoding(
            input=score,
            param_attr=fluid.ParamAttr(name='crfw')
        )

        sgd_optimizer = fluid.optimizer.AdamOptimizer(learning_rate=0.01)

        sgd_optimizer.minimize(avg_cost)

        feeder = fluid.DataFeeder(place=place, feed_list=[words, tags])

        exe = fluid.Executor(place)

        exe.run(startup_program)

        save_dirname = os.path.join(config['saved_model_dir'], str(int(time.time())))

        for pass_id in range(config['epochs']):
            print(">>> pass_id: {}".format(pass_id))
            for data in train_reader():
                # print(data)
                feed = feeder.feed(data)

                avg_loss_value, = exe.run(
                    main_program, feed=feed, fetch_list=[avg_cost],
                    return_numpy=True)
                print(avg_loss_value[0])

        if save_dirname is not None:
            fluid.io.save_inference_model(save_dirname, ['words'], [crf_decode], exe)

            # save asset
            def write_asset(output_file, data):
                os.makedirs(os.path.dirname(output_file), exist_ok=True)
                with open(output_file, 'wt') as fd:
                    fd.write('\n'.join(data))

            write_asset(os.path.join(save_dirname, 'data/vocabulary.txt'), config['vocab_data'])
            write_asset(os.path.join(save_dirname, 'data/tags.txt'), config['tags_data'])

        return None, None, save_dirname



from collections import Counter

import numpy as np


def classification_report(y_true, y_pred, labels):
    """
    Similar to the one in sklearn.metrics,
    reports per classs recall, precision and F1 score
    """
    y_true = np.asarray(y_true).ravel()
    y_pred = np.asarray(y_pred).ravel()
    corrects = Counter(yt for yt, yp in zip(y_true, y_pred) if yt == yp)
    y_true_counts = Counter(y_true)
    y_pred_counts = Counter(y_pred)
    report = ((lab,  # label
               corrects[i] / max(1, y_true_counts[i]),  # recall
               corrects[i] / max(1, y_pred_counts[i]),  # precision
               y_true_counts[i]  # support
               ) for i, lab in enumerate(labels))
    report = [(l, r, p, 2 * r * p / max(1e-9, r + p), s) for l, r, p, s in report]

    print('{:<15}{:>10}{:>10}{:>10}{:>10}\n'.format('',
                                                    'recall',
                                                    'precision',
                                                    'f1-score',
                                                    'support'))
    formatter = '{:<15}{:>10.2f}{:>10.2f}{:>10.2f}{:>10d}'.format
    for r in report:
        print(formatter(*r))
    print('')
    report2 = list(zip(*[(r * s, p * s, f1 * s) for l, r, p, f1, s in report]))
    N = len(y_true)
    print(formatter('avg / total',
                    sum(report2[0]) / N,
                    sum(report2[1]) / N,
                    sum(report2[2]) / N, N) + '\n')



import functools

import tensorflow as tf

from seq2annotation.data_input.char_level_conllz import generator_fn


# def input_fn(input_file, params=None, shuffle_and_repeat=False):
#     params = params if params is not None else {}
#     shapes = (([None], ()), [None])
#     types = ((tf.string, tf.int32), tf.string)
#     defaults = (('<pad>', 0), 'O')
#
#     dataset = tf.data.Dataset.from_generator(
#         functools.partial(generator_fn, input_file),
#         output_shapes=shapes, output_types=types)
#
#     if shuffle_and_repeat:
#         dataset = dataset.shuffle(params['buffer']).repeat(params['epochs'])
#
#     dataset = (dataset
#                .padded_batch(params.get('batch_size', 20), shapes, defaults)
#                .prefetch(1))
#
#     feature, label = dataset.make_one_shot_iterator().get_next()
#
#     # vocab_words = tf.contrib.lookup.index_table_from_file(
#     #     params['words'], num_oov_buckets=params['num_oov_buckets'])
#     #
#     # word_ids = vocab_words.lookup(feature[0])
#     #
#     # vocab_tags = tf.contrib.lookup.index_table_from_file(params['tags'])
#     # tags = vocab_tags.lookup(label)
#
#     return {'words': feature[0], 'words_len': feature[1]}, label


def index_table_from_file(vocabulary_file=None):
    index_table = {}
    index_counter = 0
    with open(vocabulary_file) as fd:
        for line in fd:
            key = line.strip()
            index_table[key] = index_counter
            index_counter += 1

    class Lookuper(object):
        def __init__(self, index_table):
            self.index_table = index_table

        def lookup(self, string):
            return self.index_table.get(string)

    return Lookuper(index_table)


def input_fn(params=None, input_file=None, config=None, shuffle_and_repeat=False):
    config = config if config is not None else {}
    shapes = (([None], ()), [None])
    types = ((tf.string, tf.int32), tf.string)
    defaults = (('<pad>', 0), 'O')

    dataset = tf.data.Dataset.from_generator(
        functools.partial(generator_fn, input_file),
        output_shapes=shapes, output_types=types)

    if shuffle_and_repeat:
        dataset = dataset.shuffle(config['buffer']).repeat(config['epochs'])

    dataset = (dataset
               .padded_batch(params['batch_size'], shapes, defaults, drop_remainder=config['use_tpu'])
               .prefetch(1))

    words_index_table = index_table_from_file(config['words'])
    tags_index_table = index_table_from_file(config['words'])
    dataset = dataset.map(lambda x, y: ((words_index_table.lookup(x[0]), x[1]), tags_index_table.lookup(y)))

    feature, label = dataset.make_one_shot_iterator().get_next()

    # vocab_words = tf.contrib.lookup.index_table_from_file(
    #     params['words'], num_oov_buckets=params['num_oov_buckets'])
    #
    # word_ids = vocab_words.lookup(feature[0])
    #
    # vocab_tags = tf.contrib.lookup.index_table_from_file(params['tags'])
    # tags = vocab_tags.lookup(label)

    return {'words': feature[0], 'words_len': feature[1]}, label


import json
import os
from collections import Counter

import numpy as np
import tensorflow as tf
from tensorflow.python.keras import models
from tensorflow.python.keras.layers import Embedding, Bidirectional, LSTM
from tensorflow.python.keras.models import Sequential
from tensorflow.python.keras import layers

from ioflow.configure import read_configure
from ioflow.corpus import get_corpus_processor
from seq2annotation.input import generate_tagset, Lookuper, \
    index_table_from_file
from tf_crf_layer.layer import CRF
from tf_crf_layer.loss import crf_loss
from tf_crf_layer.metrics import crf_accuracy
from tokenizer_tools.tagset.converter.offset_to_biluo import offset_to_biluo

config = read_configure()

corpus = get_corpus_processor(config)
corpus.prepare()
train_data_generator_func = corpus.get_generator_func(corpus.TRAIN)
eval_data_generator_func = corpus.get_generator_func(corpus.EVAL)

corpus_meta_data = corpus.get_meta_info()

raw_tag_data = corpus_meta_data['tags']
tags_data = generate_tagset(corpus_meta_data['tags'])

train_data = list(train_data_generator_func())
eval_data = list(eval_data_generator_func())

tag_lookuper = Lookuper({v: i for i, v in enumerate(tags_data)})

vocab_data_file = os.path.join(os.path.dirname(__file__), '../data/unicode_char_list.txt')
vocabulary_lookuper = index_table_from_file(vocab_data_file)


def classification_report(y_true, y_pred, labels):
    """
    Similar to the one in sklearn.metrics,
    reports per classs recall, precision and F1 score
    """
    y_true = np.asarray(y_true).ravel()
    y_pred = np.asarray(y_pred).ravel()
    corrects = Counter(yt for yt, yp in zip(y_true, y_pred) if yt == yp)
    y_true_counts = Counter(y_true)
    y_pred_counts = Counter(y_pred)
    report = ((lab,  # label
               corrects[i] / max(1, y_true_counts[i]),  # recall
               corrects[i] / max(1, y_pred_counts[i]),  # precision
               y_true_counts[i]  # support
               ) for i, lab in enumerate(labels))
    report = [(l, r, p, 2 * r * p / max(1e-9, r + p), s) for l, r, p, s in report]

    print('{:<15}{:>10}{:>10}{:>10}{:>10}\n'.format('',
                                                    'recall',
                                                    'precision',
                                                    'f1-score',
                                                    'support'))
    formatter = '{:<15}{:>10.2f}{:>10.2f}{:>10.2f}{:>10d}'.format
    for r in report:
        print(formatter(*r))
    print('')
    report2 = list(zip(*[(r * s, p * s, f1 * s) for l, r, p, f1, s in report]))
    N = len(y_true)
    print(formatter('avg / total',
                    sum(report2[0]) / N,
                    sum(report2[1]) / N,
                    sum(report2[2]) / N, N) + '\n')


def one_hot(a, num_classes):
    return np.squeeze(np.eye(num_classes)[a.reshape(-1)])


def preprocss(data, maxlen=None, intent_lookup_table=None):
    raw_x = []
    raw_y = []
    raw_intent = []

    for offset_data in data:
        tags = offset_to_biluo(offset_data)
        words = offset_data.text
        label = offset_data.extra_attr[config['intent_field']] if config['intent_field'] not in ["label"] else getattr(offset_data, config['intent_field'])

        tag_ids = [tag_lookuper.lookup(i) for i in tags]
        word_ids = [vocabulary_lookuper.lookup(i) for i in words]

        raw_x.append(word_ids)
        raw_y.append(tag_ids)
        raw_intent.append(label)

    if not intent_lookup_table:
        raw_intent_set = list(set(raw_intent))
        intent_lookup_table = Lookuper({v: i for i, v in enumerate(raw_intent_set)})

    intent_int_list = [intent_lookup_table.lookup(i) for i in raw_intent]

    if not maxlen:
        maxlen = max(len(s) for s in raw_x)

    x = tf.keras.preprocessing.sequence.pad_sequences(raw_x, maxlen,
                                                      padding='post')  # right padding

    # lef padded with -1. Indeed, any integer works as it will be masked
    # y_pos = pad_sequences(y_pos, maxlen, value=-1)
    # y_chunk = pad_sequences(y_chunk, maxlen, value=-1)
    y = tf.keras.preprocessing.sequence.pad_sequences(raw_y, maxlen, value=0,
                                                      padding='post')

    intent_np_array = np.array(intent_int_list)
    intent_one_hot = one_hot(intent_np_array, np.max(intent_np_array) + 1)

    return x, intent_one_hot, y, intent_lookup_table

MAX_LEN = 25

train_x, train_intent, train_y, intent_lookup_table = preprocss(train_data, MAX_LEN)
test_x, test_intent, test_y, _ = preprocss(eval_data, MAX_LEN, intent_lookup_table)

intent_number = train_intent.shape[1]

from tf_crf_layer.crf_dynamic_constraint_helper import generate_constraint_table, filter_constraint, sort_constraint

constraint_file = config.get("constraint")
with open(constraint_file) as fd:
    constraint = json.load(fd)

# filter out entity not in tag_list
tag_set = raw_tag_data
intent_set = intent_lookup_table.index_table.keys()
valid_constraint = filter_constraint(constraint, tag_set, intent_set)

wrong_constraint_mapping = list(valid_constraint.values())

right_constraint_mapping = sort_constraint(valid_constraint, intent_lookup_table.index_table)

tag_dict = tag_lookuper.inverse_index_table

expected_constraint_table = np.array([
    [
    #  Only allowed: Y entity for Y-domain
        #     O     B-X    B-Y    I-X    I-Y   L-X    L-Y    U-X    U-Y   start   end
        [     1,     0,     1,     0,     0,    0,     0,     0,     1,    0,     0],  # O
        [     0,     0,     0,     0,     0,    0,     0,     0,     0,    0,     0],  # B-X
        [     0,     0,     0,     0,     1,    0,     1,     0,     0,    0,     0],  # B-Y
        [     0,     0,     0,     0,     0,    0,     0,     0,     0,    0,     0],  # I-X
        [     0,     0,     0,     0,     1,    0,     1,     0,     0,    0,     0],  # I-Y
        [     0,     0,     0,     0,     0,    0,     0,     0,     0,    0,     0],  # L-X
        [     1,     0,     1,     0,     0,    0,     0,     0,     1,    0,     0],  # L-Y
        [     0,     0,     0,     0,     0,    0,     0,     0,     0,    0,     0],  # U-X
        [     1,     0,     1,     0,     0,    0,     0,     0,     1,    0,     0],  # U-Y
        [     0,     0,     0,     0,     0,    0,     0,     0,     0,    0,     0],  # start
        [     0,     0,     0,     0,     0,    0,     0,     0,     0,    0,     0],  # end
    ],
    [
    #  Only allowed: X entity for X-domain
        #     O     B-X    B-Y    I-X    I-Y   L-X    L-Y    U-X    U-Y   start   end
        [     1,     1,     0,     0,     0,    0,     0,     1,     0,    0,     0],  # O
        [     0,     0,     0,     1,     0,    1,     0,     0,     0,    0,     0],  # B-X
        [     0,     0,     0,     0,     0,    0,     0,     0,     0,    0,     0],  # B-Y
        [     0,     0,     0,     1,     0,    1,     0,     0,     0,    0,     0],  # I-X
        [     0,     0,     0,     0,     0,    0,     0,     0,     0,    0,     0],  # I-Y
        [     1,     1,     0,     0,     0,    0,     0,     1,     0,    0,     0],  # L-X
        [     0,     0,     0,     0,     0,    0,     0,     0,     0,    0,     0],  # L-Y
        [     1,     1,     0,     0,     0,    0,     0,     1,     0,    0,     0],  # U-X
        [     0,     0,     0,     0,     0,    0,     0,     0,     0,    0,     0],  # U-Y
        [     0,     0,     0,     0,     0,    0,     0,     0,     0,    0,     0],  # start
        [     0,     0,     0,     0,     0,    0,     0,     0,     0,    0,     0],  # end
    ]
], dtype=np.bool)

constraint_table = generate_constraint_table(right_constraint_mapping, tag_dict)

# diff = np.bitwise_xor(constraint_table, expected_constraint_table)

EPOCHS = 10
EMBED_DIM = 64
BiRNN_UNITS = 200

vacab_size = vocabulary_lookuper.size()
tag_size = tag_lookuper.size()

# model = Sequential()
# model.add(Embedding(vacab_size, EMBED_DIM, mask_zero=True))
# model.add(Bidirectional(LSTM(BiRNN_UNITS // 2, return_sequences=True)))
# model.add(CRF(tag_size))


raw_input = layers.Input(shape=(MAX_LEN,))
embedding_layer = Embedding(vacab_size, EMBED_DIM, mask_zero=True)(raw_input)
bilstm_layer = Bidirectional(LSTM(BiRNN_UNITS // 2, return_sequences=True))(embedding_layer)

crf_layer = CRF(
    units=tag_size,
    transition_constraint_matrix=constraint_table
)

dynamic_constraint_input = layers.Input(shape=(intent_number,))

output_layer = crf_layer([bilstm_layer, dynamic_constraint_input])

model = models.Model([raw_input, dynamic_constraint_input], output_layer)

# print model summary
model.summary()

callbacks_list = []

# tensorboard_callback = tf.keras.callbacks.TensorBoard(log_dir=config['summary_log_dir'])
# callbacks_list.append(tensorboard_callback)
#
# checkpoint_callback = tf.keras.callbacks.ModelCheckpoint(
#     os.path.join(config['model_dir'], 'cp-{epoch:04d}.ckpt'),
#     load_weights_on_restart=True,
#     verbose=1
# )
# callbacks_list.append(checkpoint_callback)

model.compile('adam', loss=crf_loss, metrics=[crf_accuracy])
model.fit(
    [train_x, train_intent], train_y,
    epochs=EPOCHS,
    validation_data=[[test_x, test_intent], test_y],
    callbacks=callbacks_list
)

# tf.keras.experimental.export_saved_model(model, config['saved_model_dir'])



import copy
import os

import tensorflow as tf

from seq2annotation import utils
from seq2annotation.utils import class_from_module_path
from tokenizer_tools.hooks import TensorObserveHook


tf.logging.set_verbosity(tf.logging.INFO)

observer_hook = TensorObserveHook(
    {"fake_golden": "fake_golden:0", "fake_prediction": "fake_prediction:0"},
    {
        # "word_str": "word_strings_Lookup:0",
        "predictions_id": "predictions:0",
        "predict_str": "predict_Lookup:0",
        "labels_id": "labels:0",
        "labels_str": "IteratorGetNext:2",
    },
    {
        "word_str": lambda x: x.decode(),
        "predict_str": lambda x: x.decode(),
        "labels_str": lambda x: x.decode(),
    },
)


def train_model(train_inpf, eval_inpf, config, model_fn, model_name):
    estimator_params = copy.deepcopy(config)

    indices = [idx for idx, tag in enumerate(config["tags_data"]) if tag.strip() != "O"]
    num_tags = len(indices) + 1
    estimator_params["_indices"] = indices
    estimator_params["_num_tags"] = num_tags

    cfg = tf.estimator.RunConfig(save_checkpoints_secs=config["save_checkpoints_secs"])

    model_specific_name = "{model_name}-{batch_size}-{learning_rate}-{max_steps}-{max_steps_without_increase}".format(
        model_name=model_name,
        batch_size=config["batch_size"],
        learning_rate=config["learning_rate"],
        max_steps=config["max_steps"],
        max_steps_without_increase=config["max_steps_without_increase"],
    )

    instance_model_dir = os.path.join(config["model_dir"], model_specific_name)

    if config["use_tpu"]:
        tpu_cluster_resolver = tf.contrib.cluster_resolver.TPUClusterResolver(
            tpu=config["tpu_name"],
            zone=config["tpu_zone"],
            project=config["gcp_project"],
        )

        run_config = tf.contrib.tpu.RunConfig(
            cluster=tpu_cluster_resolver,
            model_dir=instance_model_dir,
            session_config=tf.ConfigProto(
                allow_soft_placement=True, log_device_placement=True
            ),
            tpu_config=tf.contrib.tpu.TPUConfig(),
        )

        tpu_estimator_params = copy.deepcopy(estimator_params)
        # remove reserved keys
        # tpu_estimator_params['train_batch_size'] = tpu_estimator_params['batch_size']
        del tpu_estimator_params["batch_size"]
        # del tpu_estimator_params['context']

        estimator = tf.contrib.tpu.TPUEstimator(
            model_fn=model_fn,
            params=tpu_estimator_params,
            config=run_config,
            use_tpu=True,
            train_batch_size=estimator_params["batch_size"],
            eval_batch_size=estimator_params["batch_size"],
            predict_batch_size=estimator_params["batch_size"],
        )
    else:
        estimator = tf.estimator.Estimator(
            model_fn, instance_model_dir, cfg, estimator_params
        )

    # Path(estimator.eval_dir()).mkdir(parents=True, exist_ok=True)
    utils.create_dir_if_needed(estimator.eval_dir())

    # hook_params = params['hook']['stop_if_no_increase']
    # hook = tf.contrib.estimator.stop_if_no_increase_hook(
    #     estimator, 'f1',
    #     max_steps_without_increase=hook_params['max_steps_without_increase'],
    #     min_steps=hook_params['min_steps'],
    #     run_every_secs=hook_params['run_every_secs']
    # )

    # build hooks from config
    train_hook = []
    for i in config.get("train_hook", []):
        class_ = class_from_module_path(i["class"])
        params = i["params"]
        if i.get("inject_whole_config", False):
            params["config"] = config
        train_hook.append(class_(**params))

    eval_hook = []
    for i in config.get("eval_hook", []):
        class_ = class_from_module_path(i["class"])
        params = i["params"]
        if i.get("inject_whole_config", False):
            params["config"] = config
        eval_hook.append(class_(**params))

    if eval_inpf:
        train_spec = tf.estimator.TrainSpec(
            input_fn=train_inpf, hooks=train_hook, max_steps=config["max_steps"]
        )
        eval_spec = tf.estimator.EvalSpec(
            input_fn=eval_inpf, throttle_secs=config["throttle_secs"], hooks=eval_hook
        )
        evaluate_result, export_results = tf.estimator.train_and_evaluate(
            estimator, train_spec, eval_spec
        )
    else:
        estimator.train(
            input_fn=train_inpf, hooks=train_hook, max_steps=config["max_steps"]
        )
        evaluate_result, export_results = {}, None

        # # Write predictions to file
    # def write_predictions(name):
    #     output_file = preds_file(name)
    #     with tf.io.gfile.GFile(output_file, 'w') as f:
    #         test_inpf = functools.partial(input_fn, fwords(name))
    #         golds_gen = generator_fn(fwords(name))
    #         preds_gen = estimator.predict(test_inpf)
    #         for golds, preds in zip(golds_gen, preds_gen):
    #             ((words, _), tags) = golds
    #             preds_tags = [i.decode() for i in preds['tags']]
    #             for word, tag, tag_pred in zip(words, tags, preds_tags):
    #                 # f.write(b' '.join([word, tag, tag_pred]) + b'\n')
    #                 f.write(' '.join([word, tag, tag_pred]) + '\n')
    #             # f.write(b'\n')
    #             f.write('\n')
    #
    # for name in ['train', 'test']:
    #     write_predictions(name)

    # export saved_model
    feature_spec = {
        # 'words': tf.placeholder(tf.int32, [None, None]),
        "words": tf.placeholder(tf.string, [None, None]),
        "words_len": tf.placeholder(tf.int32, [None]),
    }

    if config.get("forced_saved_model_dir"):
        instance_saved_dir = config.get("forced_saved_model_dir")
    else:
        instance_saved_dir = os.path.join(
            config["saved_model_dir"], model_specific_name
        )

    utils.create_dir_if_needed(instance_saved_dir)

    serving_input_receiver_fn = tf.estimator.export.build_raw_serving_input_receiver_fn(
        feature_spec
    )
    raw_final_saved_model = estimator.export_saved_model(
        instance_saved_dir,
        serving_input_receiver_fn,
        # assets_extra={
        #     'tags.txt': 'data/tags.txt',
        #     'vocab.txt': 'data/unicode_char_list.txt'
        # }
    )

    final_saved_model = raw_final_saved_model.decode("utf-8")

    return evaluate_result, export_results, final_saved_model



from tensorflow.python import keras

from seq2annotation.input import Lookuper
from seq2annotation.server.inference.inference_base import InferenceBase


class KerasInferenceBase(InferenceBase):
    def __init__(self, model_path, tag_lookup_file=None, vocabulary_lookup_file=None):
        self.tag_lookup_table = Lookuper.load_from_file(tag_lookup_file)
        self.vocabulary_lookup_table = Lookuper.load_from_file(vocabulary_lookup_file)

        super(KerasInferenceBase, self).__init__(model_path)

    def preprocess(self, input_text: str):
        id_sequences = self.vocabulary_lookup_table.lookup_list_of_str_list(input_text)

        sentence = keras.preprocessing.sequence.pad_sequences(
            id_sequences, dtype="object", padding="post", truncating="post", value=0
        )

        return sentence

    def encode_input_feature(self, input_):
        return input_

    def decode_output_feature(self, input_):
        return input_

    def postprocess(self, tags_id_list):
        tags_list = self.tag_lookup_table.inverse_lookup_list_of_id_list(tags_id_list)

        return tags_list



from typing import List

import keras
import tensorflow as tf

from seq2annotation.input import Lookuper
from tokenizer_tools.tagset.NER.BILUO import BILUOSequenceEncoderDecoder
from tokenizer_tools.tagset.offset.sequence import Sequence
from tensorflow.contrib import predictor


from tokenizer_tools.tagset.exceptions import TagSetDecodeError

decoder = BILUOSequenceEncoderDecoder()

from tf_crf_layer.metrics.crf_accuracy import crf_accuracy
from tf_crf_layer.metrics.sequence_span_accuracy import sequence_span_accuracy


class Inference(object):
    def __init__(self, model_path, tag_lookup_file=None, vocabulary_lookup_file=None):
        # load model
        self.model_dir = model_path

        self.model = tf.keras.experimental.load_from_saved_model(model_path, custom_objects={"crf_accuracy": crf_accuracy, "sequence_span_accuracy": sequence_span_accuracy})
        self.predict_fn = self.model.predict

        self.tag_lookup_table = Lookuper.load_from_file(tag_lookup_file)
        self.vocabulary_lookup_table = Lookuper.load_from_file(vocabulary_lookup_file)

    def infer(self, input_text: str):
        infer_result = self._infer(input_text)
        return infer_result[0]

    def batch_infer(self, input_text: List[str]):
        return self._infer(input_text)

    def _infer(self, input_text):
        if isinstance(input_text, str):
            input_list = [input_text]
        else:
            input_list = input_text

        raw_sequences = [[i for i in text] for text in input_list]

        id_sequences = self.vocabulary_lookup_table.lookup_list_of_str_list(raw_sequences)

        sentence = keras.preprocessing.sequence.pad_sequences(
            id_sequences, dtype='object',
            padding='post', truncating='post', value=0
        )

        tags_id_list = self.predict_fn(sentence)

        tags_list = self.tag_lookup_table.inverse_lookup_list_of_id_list(tags_id_list)

        infer_result = []
        for raw_input_text, raw_text, normalized_text, tags in zip(input_list, raw_sequences, sentence, tags_list):
            # decode Unicode
            tags_seq = [i for i in tags]

            # print(tags_seq)

            # BILUO to offset
            failed = False
            try:
                seq = decoder.to_offset(tags_seq, raw_text)
            except TagSetDecodeError as e:
                print(e)

                # invalid tag sequence will raise exception
                # so return a empty result
                seq = Sequence(input_text)
                failed = True

            infer_result.append((raw_input_text, seq, tags_seq, failed))

        return infer_result


if __name__ == "__main__":
    pass



import keras
from tensorflow.contrib import predictor

from seq2annotation.server.inference.inference_base import InferenceBase
from tokenizer_tools.tagset.NER.BILUO import BILUOSequenceEncoderDecoder

decoder = BILUOSequenceEncoderDecoder()


class TensorFlowInference(InferenceBase):
    def __init__(self, *args, **kwargs):
        super(TensorFlowInference, self).__init__(*args, **kwargs)

        self.raw_sequence = None

    def instance_predict_fn(self):
        return predictor.from_saved_model(self.model_path)

    def preprocess(self, input_):
        self.raw_sequence = input_

        sentence = keras.preprocessing.sequence.pad_sequences(
            input_, dtype="object", padding="post", truncating="post", value=["<pad>"]
        ).tolist()

        return sentence

    def encode_input_feature(self, msg_list):
        input_feature = {
            "words": [[i for i in text] for text in msg_list],
            "words_len": [len(text) for text in self.raw_sequence],
        }

        return input_feature

    def decode_output_feature(self, response):
        tags_list = response["tags"].tolist()

        return [[i.decode() for i in j] for j in tags_list]

    def postprocess(self, input_):
        return input_



import json
import os
from collections import Counter

import numpy as np
import tensorflow as tf
from tensorflow.python.keras import models
from tensorflow.python.keras.layers import Embedding, Bidirectional, LSTM
from tensorflow.python.keras.models import Sequential
from tensorflow.python.keras import layers

from ioflow.configure import read_configure
from ioflow.corpus import get_corpus_processor
from seq2annotation.input import generate_tagset, Lookuper, \
    index_table_from_file
from tf_crf_layer.crf_helper import allowed_transitions
from tf_crf_layer.layer import CRF
from tf_crf_layer.loss import crf_loss
from tf_crf_layer.metrics import crf_accuracy
from tokenizer_tools.tagset.converter.offset_to_biluo import offset_to_biluo

config = read_configure()

corpus = get_corpus_processor(config)
corpus.prepare()
train_data_generator_func = corpus.get_generator_func(corpus.TRAIN)
eval_data_generator_func = corpus.get_generator_func(corpus.EVAL)

corpus_meta_data = corpus.get_meta_info()

tags_data = generate_tagset(corpus_meta_data['tags'])

train_data = list(train_data_generator_func())
eval_data = list(eval_data_generator_func())

tag_lookuper = Lookuper({v: i for i, v in enumerate(tags_data)})

vocab_data_file = os.path.join(os.path.dirname(__file__), '../data/unicode_char_list.txt')
vocabulary_lookuper = index_table_from_file(vocab_data_file)


def classification_report(y_true, y_pred, labels):
    """
    Similar to the one in sklearn.metrics,
    reports per classs recall, precision and F1 score
    """
    y_true = np.asarray(y_true).ravel()
    y_pred = np.asarray(y_pred).ravel()
    corrects = Counter(yt for yt, yp in zip(y_true, y_pred) if yt == yp)
    y_true_counts = Counter(y_true)
    y_pred_counts = Counter(y_pred)
    report = ((lab,  # label
               corrects[i] / max(1, y_true_counts[i]),  # recall
               corrects[i] / max(1, y_pred_counts[i]),  # precision
               y_true_counts[i]  # support
               ) for i, lab in enumerate(labels))
    report = [(l, r, p, 2 * r * p / max(1e-9, r + p), s) for l, r, p, s in report]

    print('{:<15}{:>10}{:>10}{:>10}{:>10}\n'.format('',
                                                    'recall',
                                                    'precision',
                                                    'f1-score',
                                                    'support'))
    formatter = '{:<15}{:>10.2f}{:>10.2f}{:>10.2f}{:>10d}'.format
    for r in report:
        print(formatter(*r))
    print('')
    report2 = list(zip(*[(r * s, p * s, f1 * s) for l, r, p, f1, s in report]))
    N = len(y_true)
    print(formatter('avg / total',
                    sum(report2[0]) / N,
                    sum(report2[1]) / N,
                    sum(report2[2]) / N, N) + '\n')


def one_hot(a, num_classes):
    return np.squeeze(np.eye(num_classes)[a.reshape(-1)])


def preprocss(data, maxlen=None, intent_lookup_table=None):
    raw_x = []
    raw_y = []
    raw_intent = []

    for offset_data in data:
        tags = offset_to_biluo(offset_data)
        words = offset_data.text
        # label = offset_data.extra_attr['domain']

        tag_ids = [tag_lookuper.lookup(i) for i in tags]
        word_ids = [vocabulary_lookuper.lookup(i) for i in words]

        raw_x.append(word_ids)
        raw_y.append(tag_ids)
        # raw_intent.append(label)

    # if not intent_lookup_table:
    #     raw_intent_set = list(set(raw_intent))
    #     intent_lookup_table = Lookuper({v: i for i, v in enumerate(raw_intent_set)})

    # intent_int_list = [intent_lookup_table.lookup(i) for i in raw_intent]

    if not maxlen:
        maxlen = max(len(s) for s in raw_x)

    x = tf.keras.preprocessing.sequence.pad_sequences(raw_x, maxlen,
                                                      padding='post')  # right padding

    # lef padded with -1. Indeed, any integer works as it will be masked
    # y_pos = pad_sequences(y_pos, maxlen, value=-1)
    # y_chunk = pad_sequences(y_chunk, maxlen, value=-1)
    y = tf.keras.preprocessing.sequence.pad_sequences(raw_y, maxlen, value=0,
                                                      padding='post')

    # intent_np_array = np.array(intent_int_list)
    # intent_one_hot = one_hot(intent_np_array, np.max(intent_np_array) + 1)
    intent_one_hot = None

    return x, intent_one_hot, y, intent_lookup_table


train_x, train_intent, train_y, intent_lookup_table = preprocss(train_data, 25)
test_x, test_intent, test_y, _ = preprocss(eval_data, 25, intent_lookup_table)

EPOCHS = 10
EMBED_DIM = 64
BiRNN_UNITS = 200

vacab_size = vocabulary_lookuper.size()
tag_size = tag_lookuper.size()

allowed = allowed_transitions("BIOUL", tag_lookuper.inverse_index_table)

model = Sequential()
model.add(Embedding(vacab_size, EMBED_DIM, mask_zero=True))
model.add(Bidirectional(LSTM(BiRNN_UNITS // 2, return_sequences=True)))
model.add(CRF(tag_size, transition_constraint=allowed))

# print model summary
model.summary()

callbacks_list = []

# tensorboard_callback = tf.keras.callbacks.TensorBoard(log_dir=config['summary_log_dir'])
# callbacks_list.append(tensorboard_callback)
#
# checkpoint_callback = tf.keras.callbacks.ModelCheckpoint(
#     os.path.join(config['model_dir'], 'cp-{epoch:04d}.ckpt'),
#     load_weights_on_restart=True,
#     verbose=1
# )
# callbacks_list.append(checkpoint_callback)

model.compile('adam', loss=crf_loss, metrics=[crf_accuracy])
model.fit(
    [train_x, train_intent], train_y,
    epochs=EPOCHS,
    validation_data=[[test_x, test_intent], test_y],
    callbacks=callbacks_list
)

# tf.keras.experimental.export_saved_model(model, config['saved_model_dir'])



import os
from collections import Counter

import numpy
import tensorflow as tf
from tensorflow.python.keras.layers import Embedding, Bidirectional, LSTM, BatchNormalization
from tensorflow.python.keras.models import Sequential

from ioflow.configure import read_configure
from ioflow.corpus import get_corpus_processor
from seq2annotation.input import generate_tagset, Lookuper, \
    index_table_from_file
from tf_crf_layer.layer import CRF
from tf_attention_layer.layers.global_attentioin_layer import GlobalAttentionLayer
from tf_crf_layer.loss import crf_loss, ConditionalRandomFieldLoss
from tf_crf_layer.metrics import crf_accuracy, SequenceCorrectness, SequenceSpanAccuracy, sequence_span_accuracy
from tokenizer_tools.tagset.converter.offset_to_biluo import offset_to_biluo

# tf.enable_eager_execution()


from seq2annotation import unrandom


config = read_configure()

corpus = get_corpus_processor(config)
corpus.prepare()
train_data_generator_func = corpus.get_generator_func(corpus.TRAIN)
eval_data_generator_func = corpus.get_generator_func(corpus.EVAL)

corpus_meta_data = corpus.get_meta_info()

tags_data = generate_tagset(corpus_meta_data['tags'])

train_data = list(train_data_generator_func())
eval_data = list(eval_data_generator_func())

tag_lookuper = Lookuper({v: i for i, v in enumerate(tags_data)})

vocab_data_file = config.get("vocabulary_file")

if not vocab_data_file:
    # load built in vocabulary file
    vocab_data_file = os.path.join(os.path.dirname(__file__), '../data/unicode_char_list.txt')

vocabulary_lookuper = index_table_from_file(vocab_data_file)


def classification_report(y_true, y_pred, labels):
    """
    Similar to the one in sklearn.metrics,
    reports per classs recall, precision and F1 score
    """
    y_true = numpy.asarray(y_true).ravel()
    y_pred = numpy.asarray(y_pred).ravel()
    corrects = Counter(yt for yt, yp in zip(y_true, y_pred) if yt == yp)
    y_true_counts = Counter(y_true)
    y_pred_counts = Counter(y_pred)
    report = ((lab,  # label
               corrects[i] / max(1, y_true_counts[i]),  # recall
               corrects[i] / max(1, y_pred_counts[i]),  # precision
               y_true_counts[i]  # support
               ) for i, lab in enumerate(labels))
    report = [(l, r, p, 2 * r * p / max(1e-9, r + p), s) for l, r, p, s in report]

    print('{:<15}{:>10}{:>10}{:>10}{:>10}\n'.format('',
                                                    'recall',
                                                    'precision',
                                                    'f1-score',
                                                    'support'))
    formatter = '{:<15}{:>10.2f}{:>10.2f}{:>10.2f}{:>10d}'.format
    for r in report:
        print(formatter(*r))
    print('')
    report2 = list(zip(*[(r * s, p * s, f1 * s) for l, r, p, f1, s in report]))
    N = len(y_true)
    print(formatter('avg / total',
                    sum(report2[0]) / N,
                    sum(report2[1]) / N,
                    sum(report2[2]) / N, N) + '\n')


def preprocss(data, maxlen=None):
    raw_x = []
    raw_y = []

    for offset_data in data:
        tags = offset_to_biluo(offset_data)
        words = offset_data.text

        tag_ids = [tag_lookuper.lookup(i) for i in tags]
        word_ids = [vocabulary_lookuper.lookup(i) for i in words]

        raw_x.append(word_ids)
        raw_y.append(tag_ids)

    if maxlen is None:
        maxlen = max(len(s) for s in raw_x)

    print(">>> maxlen: {}".format(maxlen))

    x = tf.keras.preprocessing.sequence.pad_sequences(raw_x, maxlen,
                                                      padding='post')  # right padding

    # lef padded with -1. Indeed, any integer works as it will be masked
    # y_pos = pad_sequences(y_pos, maxlen, value=-1)
    # y_chunk = pad_sequences(y_chunk, maxlen, value=-1)
    y = tf.keras.preprocessing.sequence.pad_sequences(raw_y, maxlen, value=0,
                                                      padding='post')

    return x, y


MAX_SENTENCE_LEN = config.get('max_sentence_len', 25)

train_x, train_y = preprocss(train_data, MAX_SENTENCE_LEN)
test_x, test_y = preprocss(eval_data, MAX_SENTENCE_LEN)

EPOCHS = config['epochs']
EMBED_DIM = config['embedding_dim']
BiRNN_UNITS = config['lstm_size']
USE_ATTENTION_LAYER = config.get("use_attention_layer", False)
BiLSTM_STACK_CONFIG = config.get("bilstm_stack_config", [])
BATCH_NORMALIZATION_AFTER_EMBEDDING_CONFIG = config.get("use_batch_normalization_after_embedding", False)
BATCH_NORMALIZATION_AFTER_BILSTM_CONFIG = config.get("use_batch_normalization_after_bilstm", False)
CRF_PARAMS = config.get("crf_params", {})

vacab_size = vocabulary_lookuper.size()
tag_size = tag_lookuper.size()

model = Sequential()

model.add(Embedding(vacab_size, EMBED_DIM, mask_zero=True, input_length=MAX_SENTENCE_LEN))

if BATCH_NORMALIZATION_AFTER_EMBEDDING_CONFIG:
    model.add(BatchNormalization())

for bilstm_config in BiLSTM_STACK_CONFIG:
    model.add(Bidirectional(LSTM(return_sequences=True, **bilstm_config)))

if BATCH_NORMALIZATION_AFTER_BILSTM_CONFIG:
    model.add(BatchNormalization())

if USE_ATTENTION_LAYER:
    model.add(GlobalAttentionLayer())

model.add(CRF(tag_size, name='crf', **CRF_PARAMS))

# print model summary
model.summary()

callbacks_list = []

tensorboard_callback = tf.keras.callbacks.TensorBoard(log_dir=config['summary_log_dir'])
callbacks_list.append(tensorboard_callback)

checkpoint_callback = tf.keras.callbacks.ModelCheckpoint(
    os.path.join(config['model_dir'], 'cp-{epoch:04d}.ckpt'),
    load_weights_on_restart=True,
    verbose=1
)
callbacks_list.append(checkpoint_callback)

metrics_list = []

metrics_list.append(crf_accuracy)
metrics_list.append(SequenceCorrectness())
metrics_list.append(sequence_span_accuracy)

loss_func = ConditionalRandomFieldLoss()
# loss_func = crf_loss

model.compile('adam', loss={'crf': loss_func}, metrics=metrics_list)
model.fit(
    train_x, train_y,
    epochs=EPOCHS,
    validation_data=[test_x, test_y],
    callbacks=callbacks_list
)

# Save the model
model.save(config['h5_model_file'])
tag_lookuper.dump_to_file('./results/h5_model/tag_lookup_table.json')
vocabulary_lookuper.dump_to_file('./results/h5_model/vocabulary_lookup_table.json')
tf.keras.experimental.export_saved_model(model, config['saved_model_dir'])



import sys

from flask import Flask, request, jsonify
from flask_cors import CORS

from seq2annotation.server.tensorflow_inference import Inference

app = Flask(__name__)
app.config['JSON_AS_ASCII'] = False
# app.config['DEBUG'] = True
CORS(app)

server = None


def load_predict_fn(export_dir):
    global server
    server = Inference(export_dir)

    return server


def sequence_to_response(text, seq):
    response = {
        'text': text,
        'spans': [{'start': i.start, 'end': i.end, 'type': i.entity}
                  for i in seq.span_set],
        'ents': list({i.entity.lower() for i in seq.span_set})
    }

    return response


@app.route("/parse", methods=['GET'])
def single_tokenizer():
    text_msg = request.args.get('q')

    print(text_msg)

    raw_input_text, seq, tags, failed = server.infer(text_msg)

    print(tags)
    # print(seq)

    response = sequence_to_response(raw_input_text, seq)

    return jsonify(response)


@app.route("/parse", methods=['POST'])
def batch_infer():
    text_msg = request.get_json()

    print(text_msg)

    infer_results = server.batch_infer(text_msg)

    response = []
    for raw_input_text, seq, tags, failed in infer_results:
        response.append(
            sequence_to_response(raw_input_text, seq)
        )

    return jsonify(response)


if __name__ == "__main__":
    load_predict_fn(sys.argv[1])

    app.run(host='0.0.0.0', port=5000)



import collections
from typing import List, Dict

from tokenizer_tools.tagset.offset.corpus import Corpus


def generate_constraint(
    corpus: Corpus,
    output_attr: str = None
) -> Dict[str, List[str]]:
    domain_list = collections.defaultdict(set)
    for item in corpus:
        if output_attr == "label":
            domain = item.label
        else:
            domain = item.extra_attr[output_attr]

        entity_list = []
        for span in item.span_set:
            entity = span.entity
            entity_list.append(entity)

        domain_list[domain].update(entity_list)

    domain_mapping = dict()

    # sort domain for stable output, more easy to compare and debug
    sorted_domain_list = sorted(domain_list.items(), key=lambda x: x[0])

    for k, v in sorted_domain_list:
        # sort entity for stable output, more easy to compare and debug
        sorted_entity_list = list(sorted(v))

        domain_mapping[k] = sorted_entity_list

    return domain_mapping



#!/usr/bin/env python
from typing import List

from tokenizer_tools.conllz.tag_collector import collect_label_to_file


def collect_label_to_file(input_file_list: List[str], output_file: str):
    collect_label_to_file(input_file_list, output_file)



import tensorflow as tf
from seq2annotation.algorithms.model import Model


class IdcnnCrfModel(Model):
    @classmethod
    def default_params(cls):
        default_params = {
            'filter_width': 2,
            'num_filter': 100,
            'repeat_times': 4,
            'layers': [
                {
                    'dilation': 1,
                },
                {
                    'dilation': 1,
                },
                {
                    'dilation': 2,
                }
            ]
        }

        return default_params

    def idcnn_layer(self, embeddings):
        with tf.variable_scope("idcnn"):
            model_inputs = tf.expand_dims(embeddings, 1)

            shape = [1, self.params['filter_width'], self.params['dim'], self.params['num_filter']]
            print(shape)
            filter_weights = tf.get_variable(
                "idcnn_filter",
                shape=shape)

            """
            shape of input = [batch, in_height, in_width, in_channels]
            shape of filter = [filter_height, filter_width, in_channels, out_channels]
            """
            layerInput = tf.nn.conv2d(model_inputs,
                                      filter_weights,
                                      strides=[1, 1, 1, 1],
                                      padding="SAME",
                                      name="init_layer")
            finalOutFromLayers = []
            totalWidthForLastDim = 0
            for j in range(self.params['repeat_times']):
                for i in range(len(self.params['layers'])):
                    dilation = self.params['layers'][i]['dilation']
                    isLast = True if i == (
                                len(self.params['layers']) - 1) else False
                    with tf.variable_scope("atrous-conv-layer-%d" % i,
                                           reuse=tf.AUTO_REUSE):
                        w = tf.get_variable(
                            "filterW",
                            shape=[1, self.params['filter_width'],
                                   self.params['num_filter'], self.params['num_filter']],
                            initializer=tf.contrib.layers.xavier_initializer())
                        b = tf.get_variable("filterB",
                                            shape=[self.params['num_filter']])
                        conv = tf.nn.atrous_conv2d(layerInput,
                                                   w,
                                                   rate=dilation,
                                                   padding="SAME")
                        conv = tf.nn.bias_add(conv, b)
                        conv = tf.nn.relu(conv)
                        if isLast:
                            finalOutFromLayers.append(conv)
                            totalWidthForLastDim += self.params['num_filter']
                        layerInput = conv
            finalOut = tf.concat(axis=3, values=finalOutFromLayers)
            finalOut = tf.squeeze(finalOut, [1])

        return finalOut

    def call(self, embeddings, nwords):
        data = self.idcnn_layer(embeddings)
        return data



import functools
import time
from threading import Thread
import os

from flask import Flask


def run_function_in_background(target):
    t = Thread(target=target)
    t.setDaemon(True)
    t.start()


def http_transponder(port):
    app = Flask(__name__)

    @app.route("/")
    def main():
        return 'I am the health check transponder'

    @app.route("/ping")
    def ping():
        return 'pong'

    @app.route("/are_you_ok")
    def are_you_ok():
        return "I'm OK"

    app.run(host='0.0.0.0', port=port)


# try get port config from environment
http_port = os.getenv('HEALTH_CHECK_TRANSPONDER_PORT', 9998)


run_health_check_transponder_in_background = functools.partial(
    run_function_in_background,
    functools.partial(http_transponder, port=http_port)
)


if __name__ == "__main__":
    run_health_check_transponder_in_background()

    for _ in range(100):
        time.sleep(1)
        print('.', end='', flush=True)



# below codes are copied directly from
# https://github.com/guillaumegenthial/tf_metrics/blob/master/tf_metrics/__init__.py
# since tf_metrics is not a package, can not be used as dependency,
# this is the reason why include the code directly
# all credit belongs to Guillaume Genthial

"""Multiclass"""

__author__ = "Guillaume Genthial"

import numpy as np
import tensorflow as tf
from tensorflow.python.ops.metrics_impl import _streaming_confusion_matrix


def precision(labels, predictions, num_classes, pos_indices=None,
              weights=None, average='micro'):
    """Multi-class precision metric for Tensorflow

    Parameters
    ----------
    labels : Tensor of tf.int32 or tf.int64
        The true labels
    predictions : Tensor of tf.int32 or tf.int64
        The predictions, same shape as labels
    num_classes : int
        The number of classes
    pos_indices : list of int, optional
        The indices of the positive classes, default is all
    weights : Tensor of tf.int32, optional
        Mask, must be of compatible shape with labels
    average : str, optional
        'micro': counts the total number of true positives, false
            positives, and false negatives for the classes in
            `pos_indices` and infer the metric from it.
        'macro': will compute the metric separately for each class in
            `pos_indices` and average. Will not account for class
            imbalance.
        'weighted': will compute the metric separately for each class in
            `pos_indices` and perform a weighted average by the total
            number of true labels for each class.

    Returns
    -------
    tuple of (scalar float Tensor, update_op)
    """
    cm, op = _streaming_confusion_matrix(
        labels, predictions, num_classes, weights)
    pr, _, _ = metrics_from_confusion_matrix(
        cm, pos_indices, average=average)
    op, _, _ = metrics_from_confusion_matrix(
        op, pos_indices, average=average)
    return (pr, op)


def recall(labels, predictions, num_classes, pos_indices=None, weights=None,
           average='micro'):
    """Multi-class recall metric for Tensorflow

    Parameters
    ----------
    labels : Tensor of tf.int32 or tf.int64
        The true labels
    predictions : Tensor of tf.int32 or tf.int64
        The predictions, same shape as labels
    num_classes : int
        The number of classes
    pos_indices : list of int, optional
        The indices of the positive classes, default is all
    weights : Tensor of tf.int32, optional
        Mask, must be of compatible shape with labels
    average : str, optional
        'micro': counts the total number of true positives, false
            positives, and false negatives for the classes in
            `pos_indices` and infer the metric from it.
        'macro': will compute the metric separately for each class in
            `pos_indices` and average. Will not account for class
            imbalance.
        'weighted': will compute the metric separately for each class in
            `pos_indices` and perform a weighted average by the total
            number of true labels for each class.

    Returns
    -------
    tuple of (scalar float Tensor, update_op)
    """
    cm, op = _streaming_confusion_matrix(
        labels, predictions, num_classes, weights)
    _, re, _ = metrics_from_confusion_matrix(
        cm, pos_indices, average=average)
    _, op, _ = metrics_from_confusion_matrix(
        op, pos_indices, average=average)
    return (re, op)


def f1(labels, predictions, num_classes, pos_indices=None, weights=None,
       average='micro'):
    return fbeta(labels, predictions, num_classes, pos_indices, weights,
                 average)


def fbeta(labels, predictions, num_classes, pos_indices=None, weights=None,
          average='micro', beta=1):
    """Multi-class fbeta metric for Tensorflow

    Parameters
    ----------
    labels : Tensor of tf.int32 or tf.int64
        The true labels
    predictions : Tensor of tf.int32 or tf.int64
        The predictions, same shape as labels
    num_classes : int
        The number of classes
    pos_indices : list of int, optional
        The indices of the positive classes, default is all
    weights : Tensor of tf.int32, optional
        Mask, must be of compatible shape with labels
    average : str, optional
        'micro': counts the total number of true positives, false
            positives, and false negatives for the classes in
            `pos_indices` and infer the metric from it.
        'macro': will compute the metric separately for each class in
            `pos_indices` and average. Will not account for class
            imbalance.
        'weighted': will compute the metric separately for each class in
            `pos_indices` and perform a weighted average by the total
            number of true labels for each class.
    beta : int, optional
        Weight of precision in harmonic mean

    Returns
    -------
    tuple of (scalar float Tensor, update_op)
    """
    cm, op = _streaming_confusion_matrix(
        labels, predictions, num_classes, weights)
    _, _, fbeta = metrics_from_confusion_matrix(
        cm, pos_indices, average=average, beta=beta)
    _, _, op = metrics_from_confusion_matrix(
        op, pos_indices, average=average, beta=beta)
    return (fbeta, op)


def safe_div(numerator, denominator):
    """Safe division, return 0 if denominator is 0"""
    numerator, denominator = tf.to_float(numerator), tf.to_float(denominator)
    zeros = tf.zeros_like(numerator, dtype=numerator.dtype)
    denominator_is_zero = tf.equal(denominator, zeros)
    return tf.where(denominator_is_zero, zeros, numerator / denominator)


def pr_re_fbeta(cm, pos_indices, beta=1):
    """Uses a confusion matrix to compute precision, recall and fbeta"""
    num_classes = cm.shape[0]
    neg_indices = [i for i in range(num_classes) if i not in pos_indices]
    cm_mask = np.ones([num_classes, num_classes])
    cm_mask[neg_indices, neg_indices] = 0
    diag_sum = tf.reduce_sum(tf.diag_part(cm * cm_mask))

    cm_mask = np.ones([num_classes, num_classes])
    cm_mask[:, neg_indices] = 0
    tot_pred = tf.reduce_sum(cm * cm_mask)

    cm_mask = np.ones([num_classes, num_classes])
    cm_mask[neg_indices, :] = 0
    tot_gold = tf.reduce_sum(cm * cm_mask)

    pr = safe_div(diag_sum, tot_pred)
    re = safe_div(diag_sum, tot_gold)
    fbeta = safe_div((1. + beta**2) * pr * re, beta**2 * pr + re)

    return pr, re, fbeta


def metrics_from_confusion_matrix(cm, pos_indices=None, average='micro',
                                  beta=1):
    """Precision, Recall and F1 from the confusion matrix

    Parameters
    ----------
    cm : tf.Tensor of type tf.int32, of shape (num_classes, num_classes)
        The streaming confusion matrix.
    pos_indices : list of int, optional
        The indices of the positive classes
    beta : int, optional
        Weight of precision in harmonic mean
    average : str, optional
        'micro', 'macro' or 'weighted'
    """
    num_classes = cm.shape[0]
    if pos_indices is None:
        pos_indices = [i for i in range(num_classes)]

    if average == 'micro':
        return pr_re_fbeta(cm, pos_indices, beta)
    elif average in {'macro', 'weighted'}:
        precisions, recalls, fbetas, n_golds = [], [], [], []
        for idx in pos_indices:
            pr, re, fbeta = pr_re_fbeta(cm, [idx], beta)
            precisions.append(pr)
            recalls.append(re)
            fbetas.append(fbeta)
            cm_mask = np.zeros([num_classes, num_classes])
            cm_mask[idx, :] = 1
            n_golds.append(tf.to_float(tf.reduce_sum(cm * cm_mask)))

        if average == 'macro':
            pr = tf.reduce_mean(precisions)
            re = tf.reduce_mean(recalls)
            fbeta = tf.reduce_mean(fbetas)
            return pr, re, fbeta
        if average == 'weighted':
            n_gold = tf.reduce_sum(n_golds)
            pr_sum = sum(p * n for p, n in zip(precisions, n_golds))
            pr = safe_div(pr_sum, n_gold)
            re_sum = sum(r * n for r, n in zip(recalls, n_golds))
            re = safe_div(re_sum, n_gold)
            fbeta_sum = sum(f * n for f, n in zip(fbetas, n_golds))
            fbeta = safe_div(fbeta_sum, n_gold)
            return pr, re, fbeta

    else:
        raise NotImplementedError()


import numpy as np
import tensorflow as tf
import random as rn

# The below is necessary for starting Numpy generated random numbers
# in a well-defined initial state.

np.random.seed(42)

# The below is necessary for starting core Python generated random numbers
# in a well-defined state.

rn.seed(12345)

# Force TensorFlow to use single thread.
# Multiple threads are a potential source of non-reproducible results.
# For further details, see: https://stackoverflow.com/questions/42022950/

session_conf = tf.ConfigProto(intra_op_parallelism_threads=1,
                              inter_op_parallelism_threads=1)

from tensorflow.python.keras import backend as K

# The below tf.set_random_seed() will make random number generation
# in the TensorFlow backend have a well-defined initial state.
# For further details, see:
# https://www.tensorflow.org/api_docs/python/tf/set_random_seed

tf.set_random_seed(1234)

sess = tf.Session(graph=tf.get_default_graph(), config=session_conf)
K.set_session(sess)



import os

import numpy as np
import paddle.fluid as fluid

from seq2annotation.input_paddle import read_vocabulary
from tokenizer_tools.tagset.NER.BILUO import BILUOSequenceEncoderDecoder


class Inference(object):
    def __init__(self, model_path):
        # load model
        self.place = fluid.CPUPlace()
        self.exe = fluid.Executor(self.place)
        [self.inference_program, self.feed_target_names, self.fetch_targets] = fluid.io.load_inference_model(dirname=model_path, executor=self.exe)

        # load vocabulary
        self.vocabulary = read_vocabulary(os.path.join(model_path, 'data/vocabulary.txt'))
        self.tag = read_vocabulary(os.path.join(model_path, 'data/tags.txt'))

        # prepare tag set decoder
        self.decoder = BILUOSequenceEncoderDecoder()

    def infer(self, input_text):
        data = [self.vocabulary.lookup(i) for i in input_text]
        print(data)

        word = fluid.create_lod_tensor([data], [[len(data)]], self.place)

        results, = self.exe.run(
            self.inference_program,
            feed={self.feed_target_names[0]: word},
            fetch_list=self.fetch_targets,
            return_numpy=False
        )

        # translate to str list
        result = np.array(results).reshape([-1]).tolist()
        print(result)
        output_tag = [self.tag.id_to_str(i) for i in result]
        print(output_tag)

        # decode to offset
        result = self.decoder.to_offset(output_tag, input_text)

        return result


if __name__ == '__main__':
    inference = Inference('/Users/howl/PyCharmProjects/seq2annotation/results/saved_model/1557904173')

    result = inference.infer('查这几天的天气。')

    print(result)



import sys

from flask import Flask, request, jsonify
from flask_cors import CORS
from tokenizer_tools.tagset.NER.BILUO import BILUOSequenceEncoderDecoder
from hanzi_char_lookup_feature import load_trie_from_files, generate_lookup_feature


decoder = BILUOSequenceEncoderDecoder()


app = Flask(__name__)
app.config['JSON_AS_ASCII'] = False
# app.config['DEBUG'] = True
CORS(app)

from tensorflow.contrib import predictor

predict_fn = None

t = None


def load_t(t_data):
    global t
    t = load_trie_from_files(
        {'person': [t_data]}
    )


def load_predict_fn(export_dir):
    global predict_fn
    predict_fn = predictor.from_saved_model(export_dir)

    return predict_fn


@app.route("/parse", methods=['GET'])
def single_tokenizer():
    text_msg = request.args.get('q')

    print(text_msg)

    lookup_feature = generate_lookup_feature(t, text_msg, ['person'])

    input_feature = {
            'words': [[i for i in text_msg]],
            'words_len': [len(text_msg)],
            'lookup': [lookup_feature['person']]
        }

    print(input_feature)

    predictions = predict_fn(input_feature)
    print(predictions['tags'])

    tags_seq = [i.decode() for i in predictions['tags'][0]]

    offset_list = decoder.decode_to_offset(tags_seq)
    print(offset_list)

    response = {
        'text': text_msg,
        'spans': [{'start': i[0], 'end': i[1], 'type': i[2]} for i in offset_list],
        'ents': list({i[2].lower() for i in offset_list})
    }

    return jsonify(response)


if __name__ == "__main__":
    load_predict_fn(sys.argv[1])

    load_t(sys.argv[2])

    app.run(host='0.0.0.0', port=5000)



import copy

from seq2annotation.preprocess_hooks.hook_base import HookBase
from tokenizer_tools.tagset.offset.sequence import Sequence
from tokenizer_tools.tagset.offset.span import Span


class CorpusAugment(HookBase):
    punctuation = [
        "。",
        ".",
        "？",
        "?",
        "！",
        "!"
    ]

    def __call__(self, sentence):
        text = sentence.text

        # text can be a list of str
        text_str = ''.join(text)

        no_tail_text = text_str.rstrip(''.join(self.punctuation))

        result = []

        for i in self.punctuation:
            new_sentence = copy.deepcopy(sentence)
            new_sentence.text = [i for i in ''.join((no_tail_text, i))]

            result.append(new_sentence)

        # add no punctuation one
        new_sentence = copy.deepcopy(sentence)
        new_sentence.text = [i for i in no_tail_text]
        result.append(new_sentence)

        return result[-1]


if __name__ == "__main__":
    seq = Sequence("王小明在北京的清华大学读书。")
    seq.span_set.append(Span(0, 3, 'PERSON', '王小明'))
    seq.span_set.append(Span(4, 6, 'GPE', '北京'))
    seq.span_set.append(Span(7, 11, 'ORG', '清华大学'))

    ca = CorpusAugment()
    res = ca(seq)

    print(res)



import pkg_resources

import numpy as np

from seq2annotation.trainer.train_model import train_model
from seq2annotation.algorithms.BiLSTM_CRF_model import BilstmCrfModel

from seq2annotation import utils


class Model(object):
    def __init__(self, config):
        self.config = config

    def train_and_eval_then_save(self, train_inpf, eval_inpf, configure):
        evaluate_result, export_results, final_saved_model = train_model(
            train_inpf=train_inpf,
            eval_inpf=eval_inpf,
            config=configure,
            model_fn=BilstmCrfModel.model_fn,
            model_name=BilstmCrfModel.get_model_name(),
        )

        return evaluate_result, export_results, final_saved_model

    def get_default_config(self):
        data_dir = self.config.pop("data_dir", ".")
        result_dir = self.config.pop("result_dir", ".")

        params = {
            "dim": 300,
            "dropout": 0.5,
            "num_oov_buckets": 1,
            "epochs": None,
            "batch_size": 20,
            "buffer": 15000,
            "lstm_size": 100,
            "words": utils.join_path(data_dir, "./unicode_char_list.txt"),
            "lookup": utils.join_path(data_dir, "./lookup.txt"),
            "chars": utils.join_path(data_dir, "vocab.chars.txt"),
            "tags": utils.join_path(data_dir, "./tags.txt"),
            "glove": utils.join_path(data_dir, "./glove.npz"),
            "model_dir": utils.join_path(result_dir, "model_dir"),
            "params_log_file": utils.join_path(result_dir, "params.json"),
            "train": utils.join_path(data_dir, "{}.conllz".format("train")),
            "test": utils.join_path(data_dir, "{}.conllz".format("test")),
            "preds": {
                "train": utils.join_path(result_dir, "{}.txt".format("preds_train")),
                "test": utils.join_path(result_dir, "{}.txt".format("preds_test")),
            },
            "optimizer_params": {},
            "saved_model_dir": utils.join_path(result_dir, "saved_model"),
            "hook": {
                "stop_if_no_increase": {
                    "min_steps": 100,
                    "run_every_secs": 60,
                    "max_steps_without_increase": 20,
                }
            },
            "train_spec": {"max_steps": 5000},
            "eval_spec": {"throttle_secs": 60},
            "estimator": {"save_checkpoints_secs": 120},
            "embedding": {"vocabulary_size": 128003},
            "use_tpu": False,
            "tpu_config": {"tpu_name": None, "zone": None, "gcp_project": None},
            "save_checkpoints_secs": 60,
            "learning_rate": 0.001,
            "max_steps": None,
            "max_steps_without_increase": 1000,
            "train_hook": {},
            "shuffle_pool_size": 30,
            "embedding_dim": 64,
        }

        vocab_data_file = self.config.get("vocabulary_file")

        if not vocab_data_file:
            # no vocabulary file provided, use internal one
            vocab_data_file = pkg_resources.resource_filename(
                __name__, "./data/unicode_char_list.txt"
            )
        params["vocab_data"] = np.loadtxt(
            vocab_data_file, dtype=np.unicode, comments=None, encoding=None
        ).tolist()
        params["embedding_vocabulary_size"] = len(params["vocab_data"])

        params.update(BilstmCrfModel.default_params())

        return params



import sys

from flask import Flask, request, jsonify
from flask_cors import CORS

from seq2annotation.server.paddle_inference import Inference

app = Flask(__name__)
app.config['JSON_AS_ASCII'] = False
# app.config['DEBUG'] = True
CORS(app)

server = None


def load_predict_fn(export_dir):
    global server
    server = Inference(export_dir)

    return server


@app.route("/parse", methods=['GET'])
def single_tokenizer():
    text_msg = request.args.get('q')

    print(text_msg)

    seq = server.infer(text_msg)

    response = {
        'text': text_msg,
        'spans': [{'start': i.start, 'end': i.end, 'type': i.entity} for i in
                  seq.span_set],
        'ents': list({i.entity.lower() for i in seq.span_set})
    }

    return jsonify(response)


if __name__ == "__main__":
    load_predict_fn(sys.argv[1])

    app.run(host='0.0.0.0', port=5000)



import functools

import tensorflow as tf

from seq2annotation.data_input.char_and_ngrams import generator_fn


def input_fn(input_file, params=None, shuffle_and_repeat=False):
    params = params if params is not None else {}
    shapes = (([None], (), [None, 6]), [None])
    types = ((tf.string, tf.int32, tf.float32), tf.string)
    defaults = (('<pad>', 0, 0.0), 'O')

    dataset = tf.data.Dataset.from_generator(
        functools.partial(generator_fn, input_file, params),
        output_shapes=shapes, output_types=types)

    if shuffle_and_repeat:
        dataset = dataset.shuffle(params['buffer']).repeat(params['epochs'])

    dataset = (dataset
               .padded_batch(params.get('batch_size', 20), shapes, defaults)
               .prefetch(1))

    feature, label = dataset.make_one_shot_iterator().get_next()

    # vocab_words = tf.contrib.lookup.index_table_from_file(
    #     params['words'], num_oov_buckets=params['num_oov_buckets'])
    #
    # word_ids = vocab_words.lookup(feature[0])
    #
    # vocab_tags = tf.contrib.lookup.index_table_from_file(params['tags'])
    # tags = vocab_tags.lookup(label)

    return {'words': feature[0], 'words_len': feature[1], 'lookup': feature[2]}, label


if __name__ == '__main__':
    # tf.enable_eager_execution()
    # tfe = tf.contrib.eager

    print(tf.executing_eagerly())
    with tf.Session() as sess:
        data, label = input_fn('/Users/howl/PyCharmProjects/seq2annotation_ngram_lookup_ner_on_people_daily/data/test.txt', {'trie_data_mapping': {'person': ['/Users/howl/PyCharmProjects/hanzi_char_lookup_feature/data/THUOCL_lishimingren.txt']}})
        for k, v in data.items():
            print(k)
            data = sess.run(v)
            print(data)



import os
import pathlib
import shutil
from typing import Text, Any, Type


def remove_files_in_dir(data_dir):
    input_file_list = [i.absolute() for i in pathlib.Path(data_dir).iterdir() if i.is_file()]
    for i in input_file_list:
        os.remove(i)


def remove_content_in_dir(data_dir):
    input_file_list = pathlib.Path(data_dir).iterdir()
    for i in input_file_list:
        file_path = str(i.absolute())
        if i.is_dir():
            shutil.rmtree(file_path)
        else:
            os.remove(file_path)


def create_dir_if_needed(directory):
    # copied from https://stackoverflow.com/questions/273192/how-can-i-safely-create-a-nested-directory-in-python

    # if not os.path.exists(directory):
    if not os.path.exists(directory):
        # os.makedirs(directory)
        os.makedirs(directory)

    return directory


def join_path(a, b):
    return os.path.join(a, str(pathlib.PurePosixPath(b)))


def class_from_module_path(module_path: Text) -> Type[Any]:
    # copied from rasa_nlu (https://github.com/RasaHQ/rasa) @ rasa_nlu/utils/__init__.py
    """Given the module name and path of a class, tries to retrieve the class.

    The loaded class can be used to instantiate new objects. """
    import importlib

    # load the module, will raise ImportError if module cannot be loaded
    if "." in module_path:
        module_name, _, class_name = module_path.rpartition('.')
        m = importlib.import_module(module_name)
        # get the class, will raise AttributeError if class cannot be found
        return getattr(m, class_name)
    else:
        return globals()[module_path]
    

def load_hook(hook_config):
    hook_instances = []
    for i in hook_config:
        class_ = class_from_module_path(i['class'])
        hook_instances.append(class_(**i.get('params', {})))
        
    return hook_instances



from tokenizer_tools.conll.reader import read_conll
from hanzi_char_lookup_feature.n_gram_lookup.load_dicts_from_files import load_dicts_from_files
from hanzi_char_lookup_feature.n_gram_lookup.ngrams_feature import ngrams_feature_mapping
from hanzi_char_lookup_feature.n_gram_lookup.ngrams_feature import generate_lookup_feature, load_data_set


def parse_fn(word_tag_pairs, t, params):
    # Encode in Bytes for TF
    words = [i[0] for i in word_tag_pairs]
    tags = [i[1] for i in word_tag_pairs]
    assert len(words) == len(tags), "Words and tags lengths don't match"

    words_char = ''.join(words)

    lookup_feature = generate_lookup_feature(words_char, 4, t, ['person'], dropout_rate=params['dropout_rate'])

    return (words, len(words), lookup_feature), tags


def generator_fn(input_file, params):
    # with Path(words).open('r') as f_words, Path(tags).open('r') as f_tags:
    #     for line_words, line_tags in zip(f_words, f_tags):
    #         yield parse_fn(line_words, line_tags)

    t = load_data_set(params['trie_data_mapping'])

    sentence_list = read_conll(input_file, sep=None)
    for sentence in sentence_list:
        # Encode in Bytes for TF
        word_tag_pairs = [(i[0], i[1]) for i in sentence]

        yield parse_fn(word_tag_pairs, t, params)


if __name__ == "__main__":
    # for i in generator_fn('data/corpus/train.txt'):
    #     print(i)

    for i in generator_fn('/Users/howl/PyCharmProjects/seq2annotation/data/test.txt', {'trie_data_mapping': {'person': ['/Users/howl/PyCharmProjects/hanzi_char_lookup_feature/data/THUOCL_lishimingren.txt']}}):
        print(i)



import tensorflow as tf

from seq2annotation.server.inference.keras_inference_base import KerasInferenceBase
from tf_crf_layer.metrics import crf_accuracy, sequence_span_accuracy


class TensorFlowKerasH5Inference(KerasInferenceBase):
    def instance_predict_fn(self):
        model = tf.keras.models.load_model(
            self.model_path,
            custom_objects={
                "crf_accuracy": crf_accuracy,
                "sequence_span_accuracy": sequence_span_accuracy,
            },
        )
        return model.predict



import json

from seq2annotation.helper.generate_constraint import generate_constraint
from tokenizer_tools.tagset.offset.corpus import Corpus


def generate_constraint_to_file(
    input_file: str,
    output_file: str,
    output_attr: str = "label"
):
    corpus = Corpus.read_from_file(input_file)

    domain_mapping = generate_constraint(corpus, output_attr)

    with open(output_file, "wt") as fd:
        json.dump(domain_mapping, fd, indent=4, ensure_ascii=False)



import functools

from seq2annotation.utils import load_hook
from tokenizer_tools.tagset.converter.offset_to_biluo import offset_to_biluo
from tokenizer_tools.tagset.NER.BILUO import BILUOEncoderDecoder


def generator_func(data_generator_func, config, vocabulary_lookup, tag_lookup):
    # load plugin
    preprocess_hook = load_hook(config.get('preprocess_hook', []))

    for sentence in data_generator_func():
        for hook in preprocess_hook:
            sentence = hook(sentence)

        if isinstance(sentence, list):
            for s in sentence:
                yield parse_fn(s, vocabulary_lookup, tag_lookup)
        else:
            yield parse_fn(sentence, vocabulary_lookup, tag_lookup)


def parse_fn(offset_data, vocabulary_lookup, tag_lookup):
    tags = offset_to_biluo(offset_data)
    words = offset_data.text
    assert len(words) == len(tags), "Words and tags lengths don't match"

    words_id = [vocabulary_lookup.lookup(i) for i in words]
    tags_id = [tag_lookup.lookup(i) for i in tags]

    return words_id, tags_id


class Vocabulary(object):
    def __init__(self, lookup_table):
        self.lookup_table = lookup_table
        self.reverse_lookup_table = {v: k for k, v in lookup_table.items()}

    def lookup(self, str_):
        if str_ in self.lookup_table:
            return self.lookup_table[str_]
        else:
            # not in the table: return extra max(id) + 1
            return len(self.lookup_table)

    def length(self):
        return len(self.lookup_table)

    def id_to_str(self, id_):
        if id_ in self.reverse_lookup_table:
            return self.reverse_lookup_table[id_]
        else:
            return '<UNK>'


def read_vocabulary(vocabulary):
    data = {}

    fd = open(vocabulary) if isinstance(vocabulary, str) else vocabulary
    for line in fd:
        word = line.strip()
        data[word] = len(data)

    if isinstance(vocabulary, str):
        fd.close()

    return Vocabulary(data)


def build_input_func(data_generator_func, config=None):
    vocabulary_lookup = read_vocabulary(config['vocab_data'])
    tag_lookup = read_vocabulary(config['tags_data'])

    def input_func():
        train_dataset = generator_func(data_generator_func, config, vocabulary_lookup, tag_lookup)

        return train_dataset
    
    return input_func


def build_gold_generator_func(offset_dataset):
    return functools.partial(generator_func, offset_dataset)


def generate_tagset(tags):
    if not tags:
        return []

    tagset = set()
    for tag in tags:
        encoder = BILUOEncoderDecoder(tag)
        tagset.update(encoder.all_tag_set())

    tagset_list = list(tagset)
    
    # make sure O is first tag,
    # this is a bug feature, otherwise sentence_correct is not correct
    # due to the crf decoder, need fix
    tagset_list.remove(BILUOEncoderDecoder.oscar)
    tagset_list.insert(0, BILUOEncoderDecoder.oscar)

    return tagset_list




import pkg_resources

import numpy as np

from seq2annotation.trainer.train_model_paddle import train_model
from seq2annotation import utils


class Model(object):
    def __init__(self, config):
        self.config = config

    def train_and_eval_then_save(self, train_inpf, eval_inpf, configure):
        evaluate_result, export_results, final_saved_model = train_model(
            train_inpf=train_inpf,
            eval_inpf=eval_inpf,
            config=configure
        )

        return evaluate_result, export_results, final_saved_model

    def get_default_config(self):
        data_dir = self.config.pop('data_dir', '.')
        result_dir = self.config.pop('result_dir', '.')

        params = {
            'dim': 300,
            'dropout': 0.5,
            'num_oov_buckets': 1,
            'epochs': 10,
            'batch_size': 20,
            'buffer': 15000,
            'lstm_size': 100,
            'words': utils.join_path(data_dir, './unicode_char_list.txt'),
            'lookup': utils.join_path(data_dir, './lookup.txt'),
            'chars': utils.join_path(data_dir, 'vocab.chars.txt'),
            'tags': utils.join_path(data_dir, './tags.txt'),
            'glove': utils.join_path(data_dir, './glove.npz'),

            'model_dir': utils.join_path(result_dir, 'model_dir'),
            'params_log_file': utils.join_path(result_dir, 'params.json'),

            'train': utils.join_path(data_dir, '{}.conllz'.format('train')),
            'test': utils.join_path(data_dir, '{}.conllz'.format('test')),

            'preds': {
                'train': utils.join_path(result_dir, '{}.txt'.format('preds_train')),
                'test': utils.join_path(result_dir, '{}.txt'.format('preds_test')),
            },

            'optimizer_params': {},

            'saved_model_dir': utils.join_path(result_dir, 'saved_model'),

            'hook': {
                'stop_if_no_increase': {
                    'min_steps': 100,
                    'run_every_secs': 60,
                    'max_steps_without_increase': 20
                }
            },

            'train_spec': {
                'max_steps': 5000
            },
            'eval_spec': {
                'throttle_secs': 60
            },

            'estimator': {
                'save_checkpoints_secs': 120
            },


            'embedding': {
                'vocabulary_size': 128003
            },

            'use_tpu': False,
            'tpu_config': {
                'tpu_name': None,
                'zone': None,
                'gcp_project': None
            },

            'save_checkpoints_secs': 60,
            'learning_rate': 0.001,
            'max_steps': None,
            'max_steps_without_increase': 1000,
            'train_hook': {},
            'shuffle_pool_size': 30,
            'embedding_dim': 64,
        }

        vocab_data_file = pkg_resources.resource_filename(__name__,
                                                          './data/unicode_char_list.txt')
        params['vocab_data'] = np.loadtxt(vocab_data_file, dtype=np.unicode,
                                          encoding=None).tolist()

        # plus one for OOV
        params['embedding_vocabulary_size'] = len(params['vocab_data']) + 1

        return params



from ioflow.corpus import Corpus
from ioflow.task_status import get_task_status_class
from ioflow.model_saver import ModelSaver
from ioflow.performance_metrics import get_performance_metrics_class
from ioflow.configure import read_configure

from seq2annotation.input_paddle import build_input_func, generate_tagset
from seq2annotation.model_paddle import Model


class Train(object):
    def train(self, addition_config=None, *args, **kwargs):
        raw_config = read_configure(*args, **kwargs)

        # add dynamic configure
        if addition_config:
            raw_config.update(addition_config)

        model = Model(raw_config)

        config = model.get_default_config()
        config.update(raw_config)

        task_status_class = get_task_status_class(config)
        task_status = task_status_class(config)

        # read data according configure
        corpus = Corpus(config)
        corpus.prepare()
        train_data_generator_func = corpus.get_generator_func(corpus.TRAIN)
        eval_data_generator_func = corpus.get_generator_func(corpus.EVAL)

        corpus_meta_data = corpus.get_meta_info()

        config['tags_data'] = generate_tagset(corpus_meta_data['tags'])

        # build model according configure

        # send START status to monitor system
        task_status.send_status(task_status.START)

        # train and evaluate model
        train_input_func = build_input_func(train_data_generator_func, config)
        eval_input_func = build_input_func(eval_data_generator_func, config)

        evaluate_result, export_results, final_saved_model = model.train_and_eval_then_save(
            train_input_func,
            eval_input_func,
            config
        )

        task_status.send_status(task_status.DONE)

        if evaluate_result:
            performance_metrics_class = get_performance_metrics_class(config)
            performance_metrics = performance_metrics_class(config)
            performance_metrics.set_metrics('test_loss',
                                            evaluate_result['loss'])
            performance_metrics.set_metrics('test_acc', evaluate_result['acc'])

        model_saver = ModelSaver(config)
        model_saver.save_model(final_saved_model)

        return final_saved_model


if __name__ == "__main__":
    train = Train()
    train.train()



import copy
import functools
import json
from pathlib import Path

import tensorflow as tf

# from seq2annotation.data_input.simple import input_fn as simple_input_fn
from seq2annotation.data_input.with_lookup import input_fn as simple_input_fn
from seq2annotation.data_input.simple import generator_fn as simple_generator_fn
from seq2annotation.algorithms.BiLSTM_CRF_model import BilstmCrfModel

tf.logging.set_verbosity(tf.logging.INFO)


def train_model(**kwargs):
    data_dir = kwargs.pop('data_dir', '.')
    result_dir = kwargs.pop('result_dir', '.')
    input_fn = kwargs.pop('input_fn', simple_input_fn)
    generator_fn = kwargs.pop('generator_fn', simple_generator_fn)
    model_fn = kwargs.pop('model_fn', BilstmCrfModel.model_fn)

    params = {
        'dim': 300,
        'dropout': 0.5,
        'num_oov_buckets': 1,
        'epochs': None,
        'batch_size': 20,
        'buffer': 15000,
        'lstm_size': 100,
        'words': str(Path(data_dir, './unicode_char_list.txt')),
        'lookup': str(Path(data_dir, './lookup.txt')),
        'chars': str(Path(data_dir, 'vocab.chars.txt')),
        'tags': str(Path(data_dir, './tags.txt')),
        'glove': str(Path(data_dir, './glove.npz')),

        'model_dir': str(Path(result_dir, 'model_dir')),
        'params_log_file': str(Path(result_dir, 'params.json')),

        'train': str(Path(data_dir, '{}.txt'.format('train'))),
        'test': str(Path(data_dir, '{}.txt'.format('test'))),

        'preds': {
            'train': str(Path(result_dir, '{}.txt'.format('preds_train'))),
            'test': str(Path(result_dir, '{}.txt'.format('preds_test'))),
        },

        'saved_model_dir': str(Path(result_dir, 'saved_model')),

        'hook': {
            'stop_if_no_increase': {
                'min_steps': 100,
                'run_every_secs': 60,
                'max_steps_without_increase': 20
            }
        },

        'train_spec': {
            'max_steps': 5000
        },
        'eval_spec': {
            'throttle_secs': 120
        },

        'estimator': {
            'save_checkpoints_secs': 120
        },


        'embedding': {
            'vocabulary_size': 128003
        }
    }

    # update from kwargs
    params.update(kwargs)

    with open(params['params_log_file'], 'w') as f:
        json.dump(params, f, indent=4, sort_keys=True)

    def fwords(name):
        return params[name]

    def preds_file(name):
        return params['preds'][name]

    # Estimator, train and evaluate
    train_inpf = functools.partial(input_fn, fwords('train'),
                                   params, shuffle_and_repeat=True)
    eval_inpf = functools.partial(input_fn, fwords('test'))

    estimator_params = copy.deepcopy(params)
    # estimator_params.update({
    #     'words_feature_columns': words_feature_columns,
    #     'words_len_feature_columns': words_len_feature_columns
    # })

    cfg = tf.estimator.RunConfig(save_checkpoints_secs=params['estimator']['save_checkpoints_secs'])

    estimator = tf.estimator.Estimator(model_fn, params['model_dir'], cfg, estimator_params)
    Path(estimator.eval_dir()).mkdir(parents=True, exist_ok=True)

    hook_params = params['hook']['stop_if_no_increase']
    hook = tf.contrib.estimator.stop_if_no_increase_hook(
        estimator, 'f1',
        max_steps_without_increase=hook_params['max_steps_without_increase'],
        min_steps=hook_params['min_steps'],
        run_every_secs=hook_params['run_every_secs']
    )

    train_spec = tf.estimator.TrainSpec(input_fn=train_inpf, hooks=[hook], max_steps=params['train_spec']['max_steps'])
    eval_spec = tf.estimator.EvalSpec(input_fn=eval_inpf, throttle_secs=params['eval_spec']['throttle_secs'])
    tf.estimator.train_and_evaluate(estimator, train_spec, eval_spec)
    # estimator.train(input_fn=train_inpf, hooks=[hook])

    # Write predictions to file
    def write_predictions(name):
        output_file = preds_file(name)
        with open(output_file, 'wt') as f:
            test_inpf = functools.partial(input_fn, fwords(name))
            golds_gen = generator_fn(fwords(name))
            preds_gen = estimator.predict(test_inpf)
            for golds, preds in zip(golds_gen, preds_gen):
                ((words, _), tags) = golds
                preds_tags = [i.decode() for i in preds['tags']]
                for word, tag, tag_pred in zip(words, tags, preds_tags):
                    # f.write(b' '.join([word, tag, tag_pred]) + b'\n')
                    f.write(' '.join([word, tag, tag_pred]) + '\n')
                # f.write(b'\n')
                f.write('\n')

    for name in ['train', 'test']:
        write_predictions(name)

    # export saved_model
    feature_spec = {
        'words': tf.placeholder(tf.string, [None, None]),
        'words_len': tf.placeholder(tf.int32, [None]),
        'lookup': tf.placeholder(tf.float32, [None, None, 6])
    }
    serving_input_receiver_fn = tf.estimator.export.build_raw_serving_input_receiver_fn(feature_spec)
    estimator.export_saved_model(
        params['saved_model_dir'],
        serving_input_receiver_fn,
    )



from ioflow.corpus import get_corpus_processor
from ioflow.eval_reporter import get_eval_reporter
from ioflow.task_status import get_task_status
from ioflow.model_saver import get_model_saver
from ioflow.configure import read_configure

from seq2annotation.input import build_input_func, generate_tagset
from seq2annotation.model import Model
from seq2annotation.health_check_transponder import (
    run_health_check_transponder_in_background,
)

# start health check
run_health_check_transponder_in_background()

raw_config = read_configure()
model = Model(raw_config)

config = model.get_default_config()
config.update(raw_config)

task_status = get_task_status(config)

task_status.send_status(task_status.START)

# read data according configure
# # report status: start to download corpus
task_status.send_status(task_status.START_DOWNLOAD_CORPUS)

corpus = get_corpus_processor(config)
corpus.prepare()
train_data_generator_func = corpus.get_generator_func(corpus.TRAIN)
eval_data_generator_func = corpus.get_generator_func(corpus.EVAL)

corpus_meta_data = corpus.get_meta_info()

config["tags_data"] = generate_tagset(corpus_meta_data["tags"])

# build model according configure


# # report status: start to process corpus
task_status.send_status(task_status.START_PROCESS_CORPUS)

# train and evaluate model
train_input_func = build_input_func(train_data_generator_func, config)
eval_input_func = (
    build_input_func(eval_data_generator_func, config)
    if eval_data_generator_func
    else None
)

# # report status: start to train
task_status.send_status(task_status.START_TRAIN)
task_status.send_progress(0)
task_status.send_progress(50)

evaluate_result, export_results, final_saved_model = model.train_and_eval_then_save(
    train_input_func, eval_input_func, config
)

task_status.send_progress(100)

# # report status: start to test
task_status.send_status(task_status.START_TEST)

from seq2annotation.server.tensorflow_inference import Inference
from seq2annotation.server.http import sequence_to_response

eval_reporter = get_eval_reporter(config)

# result_list = []
inference = Inference(final_saved_model)
for item in eval_data_generator_func():
    text, result, _, _ = inference.infer("".join(item.text))
    eval_reporter.record_x_and_y(item, sequence_to_response(text, result))
    # result_list.append(result)

eval_reporter.submit()

# # report status: start to upload model
task_status.send_status(task_status.START_UPLOAD_MODEL)

model_saver = get_model_saver(config)
model_saver.save_model(final_saved_model)

task_status.send_status(task_status.DONE)



from tokenizer_tools.conll.reader import read_conll


def parse_fn(word_tag_pairs):
    # Encode in Bytes for TF
    words = [i[0] for i in word_tag_pairs]
    tags = [i[1] for i in word_tag_pairs]
    assert len(words) == len(tags), "Words and tags lengths don't match"
    return (words, len(words)), tags


def generator_fn(input_file):
    # with Path(words).open('r') as f_words, Path(tags).open('r') as f_tags:
    #     for line_words, line_tags in zip(f_words, f_tags):
    #         yield parse_fn(line_words, line_tags)

    sentence_list = read_conll(input_file, sep=None)
    for sentence in sentence_list:
        # Encode in Bytes for TF
        word_tag_pairs = [(i[0], i[1]) for i in sentence]

        yield parse_fn(word_tag_pairs)


if __name__ == "__main__":
    # for i in generator_fn('data/corpus/train.txt'):
    #     print(i)

    for i in generator_fn('data/corpus/test.txt'):
        print(i)



import os

import setuptools
from setuptools import setup

install_requires = [
    "numpy",
    "keras",
    "tokenizer_tools",
    "flask",
    "flask-cors",
    "ioflow",
    "tf-crf-layer",
    "tf-attention-layer",
    "tensorflow>=1.15.0,<2.0.0"
]


setup(
    name=os.getenv("_PKG_NAME", "seq2annotation"),  # _PKG_NAME will be used in Makefile for dev release
    version="0.8.0",
    packages=setuptools.find_packages(),
    include_package_data=True,
    url="https://github.com/howl-anderson/seq2annotation",
    license="Apache 2.0",
    author="Xiaoquan Kong",
    author_email="u1mail2me@gmail.com",
    description="seq2annotation",
    install_requires=install_requires,
)



import collections
import functools
import json
import logging
from collections import OrderedDict
from typing import Dict, List

import tensorflow as tf

from seq2annotation.utils import class_from_module_path, load_hook
from tokenizer_tools.tagset.converter.offset_to_biluo import offset_to_biluo
from tokenizer_tools.tagset.NER.BILUO import BILUOEncoderDecoder

logger = logging.getLogger(__name__)


class Lookuper(object):
    def __init__(self, index_table: Dict[str, int]):
        # index_table: str -> int, ordered by key
        self.index_table = OrderedDict(sorted(index_table.items(), key=lambda x: x[0]))
        # inverse index table: int -> str
        self.inverse_index_table = OrderedDict(sorted(
            [(v, k) for k, v in self.index_table.items()],
            key=lambda x: x[0]
        ))  # type: OrderedDict[int, str]

    def lookup(self, string: str):
        if string not in self.index_table:
            return 1
            raise ValueError("'{}' not in index_table".format(string))
        else:
            return self.index_table.get(string)

    def lookup_str_list(self, str_list: List[str]) -> List[int]:
        return list([self.lookup(i) for i in str_list])

    def lookup_list_of_str_list(self, list_of_str_list: List[List[str]]) -> List[List[int]]:
        list_of_id_list = []
        for str_list in list_of_str_list:
            id_list = self.lookup_str_list(str_list)
            list_of_id_list.append(id_list)

        return list_of_id_list

    def inverse_lookup(self, id_: int):
        if id_ not in self.inverse_index_table:
            return 0
        else:
            return self.inverse_index_table.get(id_)

    def inverse_lookup_id_list(self, id_list: List[int]):
        return list([self.inverse_lookup(i) for i in id_list])
    
    def inverse_lookup_list_of_id_list(self, list_of_id_list: List[List[int]]):
        list_of_str_list = []
        for id_list in list_of_id_list:
            str_list = self.inverse_lookup_id_list(id_list)
            list_of_str_list.append(str_list)

        return list_of_str_list

    def size(self) -> int:
        return len(self.index_table)

    def check_id_continuity(self) -> bool:
        for i in range(self.size()):
            if i not in self.inverse_index_table:
                return False
        return True

    def tolist(self) -> List[str]:
        assert self.check_id_continuity()

        return [self.inverse_index_table[i] for i in range(self.size())]

    @classmethod
    def load_from_file(cls, data_file):
        with open(data_file, 'rt') as fd:
            # since json or yaml can not guarantee the dict order, list of (key, value) is adopted
            paired_dict = json.load(fd)

            return cls(dict(paired_dict))

    def dump_to_file(self, data_file):
        with open(data_file, 'wt') as fd:
            # since json or yaml can not guarantee the dict order, list of (key, value) is adopted
            paired_dict = list((k, v) for k, v in self.index_table.items())

            # set ensure_ascii=False for human readability of dumped file
            json.dump(paired_dict, fd, ensure_ascii=False)


def index_table_from_file(vocabulary_file=None):
    index_table = {}
    index_counter = 1
    with open(vocabulary_file) as fd:
        for line in fd:
            key = line.strip('\n')
            index_table[key] = index_counter
            index_counter += 1

    return Lookuper(index_table)


def read_assets():
    return {
        'vocab_filename': 'data/unicode_char_list.txt',
        'tag_filename': 'data/tags.txt'
    }


def generator_func(data_generator_func, config):
    # load plugin
    preprocess_hook = load_hook(config.get('preprocess_hook', []))

    for sentence in data_generator_func():
        for hook in preprocess_hook:
            sentence = hook(sentence)

        if isinstance(sentence, list):
            for s in sentence:
                yield parse_fn(s)
        else:
            yield parse_fn(sentence)


def parse_fn(offset_data):
    tags = offset_to_biluo(offset_data)
    words = offset_data.text
    assert len(words) == len(tags), "Words and tags lengths don't match"

    logger.debug((words, len(words)), tags)

    return (words, len(words)), tags


def parse_to_dataset(data_generator_func, config=None, shuffle_and_repeat=False):
    config = config if config is not None else {}
    shapes = (([None], ()), [None])
    types = ((tf.string, tf.int32), tf.string)
    defaults = (('<pad>', 0), 'O')

    dataset = tf.data.Dataset.from_generator(
        functools.partial(generator_func, data_generator_func, config),
        output_shapes=shapes, output_types=types)

    if shuffle_and_repeat:
        # print(">>> {}".format(config))
        dataset = dataset.shuffle(config['shuffle_pool_size']).repeat(config['epochs'])

    # char_encoder = tfds.features.text.SubwordTextEncoder.load_from_file(read_assets()['vocab_filename'])
    # tag_encoder = tfds.features.text.SubwordTextEncoder.load_from_file(read_assets()['tag_filename'])
    # dataset = dataset.map(lambda x: (char_encoder.encode(x[0][0]), tag_encoder.encode(x[0][1]), x[1]))

    # words_index_table = index_table_from_file(read_assets()['vocab_filename'])
    # tags_index_table = index_table_from_file(read_assets()['tag_filename'])
    # dataset = dataset.map(lambda x, y: ((words_index_table.lookup(x[0]), x[1]), tags_index_table.lookup(y)))

    dataset = (dataset
               .padded_batch(config['batch_size'], shapes, defaults)
               .prefetch(1))

    return dataset


def dataset_to_feature_column(dataset):
    (words, words_len), label = dataset.make_one_shot_iterator().get_next()

    # word_index_lookuper = tf.contrib.lookup.index_table_from_file(
    #     read_assets()['vocab_filename'],
    #     num_oov_buckets=1
    # )
    # words = word_index_lookuper.lookup(words)
    #
    # tag_index_lookuper = tf.contrib.lookup.index_table_from_file(
    #     read_assets()['tag_filename'],
    #     num_oov_buckets=1
    # )
    # label = tag_index_lookuper.lookup(label)

    return {'words': words, 'words_len': words_len}, label


def build_input_func(data_generator_func, config=None):
    def input_func():
        train_dataset = parse_to_dataset(data_generator_func, config, shuffle_and_repeat=True)
        data_iterator = dataset_to_feature_column(train_dataset)
        
        return data_iterator
    
    return input_func


def build_gold_generator_func(offset_dataset):
    return functools.partial(generator_func, offset_dataset)


def generate_tagset(tags) -> List[str]:
    if not tags:
        return []

    tagset = set()
    for tag in tags:
        encoder = BILUOEncoderDecoder(tag)
        tagset.update(encoder.all_tag_set())

    tagset_list = list(tagset)
    
    # make sure O is first tag,
    # this is a bug feature, otherwise sentence_correct is not correct
    # due to the crf decoder, need fix
    tagset_list.remove(BILUOEncoderDecoder.oscar)
    tagset_list = list(sorted(tagset_list, key=lambda x: x))

    tagset_list.insert(0, BILUOEncoderDecoder.oscar)

    return tagset_list




from typing import List, Union, Tuple

import tensorflow as tf
import numpy as np

from tokenizer_tools.tagset.converter.offset_to_biluo import offset_to_biluo
from tokenizer_tools.tagset.offset.sequence import Sequence
from seq2annotation.input import Lookuper


def str_to_id(string: Union[str, List[str]], vocabulary_look_table: Lookuper) -> List[int]:
    id_list = [vocabulary_look_table.lookup(i) for i in string]

    return id_list


def id_to_str(id_list: List[int], vocabulary_look_table: Lookuper) -> List[str]:
    str_list = [vocabulary_look_table.inverse_lookup(id) for id in id_list]

    return str_list


def preprocess(
    data: List[Sequence],
    tag_lookup_table: Lookuper,
    vocabulary_look_table: Lookuper,
    seq_maxlen: Union[None, int] = None,
) -> Tuple[np.ndarray, np.ndarray, int]:
    raw_x = []
    raw_y = []

    for offset_data in data:
        tags = offset_to_biluo(offset_data)
        words = offset_data.text

        tag_ids = [tag_lookup_table.lookup(i) for i in tags]
        word_ids = [vocabulary_look_table.lookup(i) for i in words]

        raw_x.append(word_ids)
        raw_y.append(tag_ids)

    if not seq_maxlen:
        seq_maxlen = max(len(s) for s in raw_x)

    print(">>> maxlen: {}".format(seq_maxlen))

    x = tf.keras.preprocessing.sequence.pad_sequences(
        raw_x, seq_maxlen, padding="post"
    )  # right seq_maxlen

    # lef padded with -1. Indeed, any integer works as it will be masked
    # y_pos = pad_sequences(y_pos, maxlen, value=-1)
    # y_chunk = pad_sequences(y_chunk, maxlen, value=-1)
    y = tf.keras.preprocessing.sequence.pad_sequences(
        raw_y, seq_maxlen, value=0, padding="post"
    )

    return x, y, seq_maxlen



from pathlib import Path

import tensorflow as tf
import numpy as np
from seq2annotation.metrics import precision, recall, f1
from tokenizer_tools.metrics import correct_rate


class Model(object):
    @classmethod
    def default_params(cls):
        return {}

    @classmethod
    def get_model_name(cls):
        return cls.__name__

    @classmethod
    def model_fn(cls, features, labels, mode, params):
        instance = cls(features, labels, mode, params)
        return instance()

    def __init__(self, features, labels, mode, params):
        self.features = features
        self.labels = labels
        self.mode = mode
        self.params = params

    def tpu_input_layer(self):
        word_ids = self.features["words"]

        nwords = tf.identity(self.features["words_len"], name="input_words_len")

        indices = self.params["_indices"]
        num_tags = self.params["_num_tags"]

        return indices, num_tags, word_ids, nwords

    def input_layer(self):
        # data = np.loadtxt(self.params['vocab'], dtype=np.unicode, encoding=None)
        data = self.params["vocab_data"]
        mapping_strings = tf.Variable(data)
        vocab_words = tf.contrib.lookup.index_table_from_tensor(
            mapping_strings, num_oov_buckets=1
        )

        # Word Embeddings
        words = tf.identity(self.features["words"], name="input_words")
        word_ids = vocab_words.lookup(words)

        #
        # raw_nwords = tf.identity(features['words_len'], name='input_words_len')
        # nwords = tf.feature_column.input_layer({'words_len': raw_nwords}, params['words_len_feature_columns'])
        # nwords = tf.reshape(nwords, [-1])
        # nwords = tf.to_int32(nwords)

        # words = features['words']
        # words = tf.convert_to_tensor(words)
        #
        # nwords = features['words_len']
        # nwords = tf.convert_to_tensor(nwords)

        nwords = tf.identity(self.features["words_len"], name="input_words_len")

        # get tag info
        # with Path(self.params['tags']).open() as f:
        indices = [
            idx
            for idx, tag in enumerate(self.params["tags_data"])
            if tag.strip() != "O"
        ]
        num_tags = len(indices) + 1

        # # true tags to ids
        # if self.mode == tf.estimator.ModeKeys.PREDICT:
        #     true_tag_ids = 0
        # else:
        #     true_tag_ids = self.tag2id(self.labels)

        return indices, num_tags, word_ids, nwords

    def embedding_layer(self, word_ids):
        # load pre-trained data from file
        # glove = np.load(params['glove'])['embeddings']  # np.array

        # training the embedding during training
        glove = np.zeros(
            (self.params["embedding_vocabulary_size"], self.params["embedding_dim"]),
            dtype=np.float32,
        )

        # Add OOV word embedding
        embedding_array = np.vstack([glove, [[0.0] * self.params["embedding_dim"]]])

        embedding_variable = tf.Variable(
            embedding_array, dtype=tf.float32, trainable=True
        )
        embeddings = tf.nn.embedding_lookup(embedding_variable, word_ids)

        return embeddings

    def dropout_layer(self, data):
        training = self.mode == tf.estimator.ModeKeys.TRAIN
        output = tf.layers.dropout(data, rate=self.params["dropout"], training=training)

        return output

    def dense_layer(self, data, num_tags):
        logits = tf.layers.dense(data, num_tags)

        return logits

    def load_tag_data(self):
        # data = np.loadtxt(self.params['tags'], dtype=np.unicode, encoding=None)
        data = self.params["tags_data"]
        mapping_strings = tf.Variable(data)

        return mapping_strings

    def load_word_data(self):
        data = np.loadtxt(self.params["words"], dtype=np.unicode, encoding=None)
        mapping_strings = tf.Variable(data.reshape((-1,)))

        return mapping_strings

    def tag2id(self, labels, name=None):
        mapping_strings = self.load_tag_data()
        vocab_tags = tf.contrib.lookup.index_table_from_tensor(
            mapping_strings, name=name
        )

        tags = vocab_tags.lookup(labels)

        return tags

    def id2tag(self, pred_ids, name=None):
        mapping_strings = self.load_tag_data()
        reverse_vocab_tags = tf.contrib.lookup.index_to_string_table_from_tensor(
            mapping_strings, name=name
        )

        pred_strings = reverse_vocab_tags.lookup(tf.to_int64(pred_ids))

        return pred_strings

    def id2word(self, word_ids, name=None):
        mapping_strings = self.load_word_data()
        reverse_vocab_tags = tf.contrib.lookup.index_to_string_table_from_tensor(
            mapping_strings, name=name
        )

        word_strings = reverse_vocab_tags.lookup(tf.to_int64(word_ids))

        return word_strings

    def loss_layer(self, preds, ground_true, nwords, crf_params):
        with tf.name_scope("CRF_log_likelihood"):
            log_likelihood, _ = tf.contrib.crf.crf_log_likelihood(
                preds, ground_true, nwords, crf_params
            )

        loss = tf.reduce_mean(-log_likelihood)

        return loss

    def crf_decode_layer(self, logits, crf_params, nwords):
        with tf.name_scope("CRF_decode"):
            pred_ids, _ = tf.contrib.crf.crf_decode(logits, crf_params, nwords)

        return pred_ids

    def compute_metrics(self, tags, pred_ids, num_tags, indices, nwords):
        weights = tf.sequence_mask(nwords)

        # metrics_correct_rate, golden, predict = correct_rate(tags, pred_ids)
        # metrics_correct_rate = correct_rate(tags, pred_ids, weights)

        metrics = {
            "acc": tf.metrics.accuracy(tags, pred_ids, weights),
            "precision": precision(tags, pred_ids, num_tags, indices, weights),
            "recall": recall(tags, pred_ids, num_tags, indices, weights),
            "f1": f1(tags, pred_ids, num_tags, indices, weights),
            "correct_rate": correct_rate(tags, pred_ids, weights),
            # 'golden': (golden, tf.zeros([], tf.int32)),
            # 'predict': (predict, tf.zeros([], tf.int32))
        }

        for metric_name, op in metrics.items():
            tf.summary.scalar(metric_name, op[1])

        return metrics

    def call(self, embeddings, nwords):
        raise NotImplementedError

    def __call__(self):
        indices, num_tags, word_ids, nwords = self.input_layer()
        # indices, num_tags, word_ids, nwords = self.tpu_input_layer()

        embeddings = self.embedding_layer(word_ids)

        data = self.call(embeddings, nwords)

        data = self.dropout_layer(data)
        logits = self.dense_layer(data, num_tags)

        crf_params = tf.get_variable("crf", [num_tags, num_tags], dtype=tf.float32)

        pred_ids = self.crf_decode_layer(logits, crf_params, nwords)
        pred_strings = self.id2tag(pred_ids, name="predict")

        # word_strings = self.id2word(word_ids, name='word_strings')

        # print(word_strings)

        if self.mode == tf.estimator.ModeKeys.PREDICT:
            predictions = {"pred_ids": pred_ids, "tags": pred_strings}

            if self.params["use_tpu"]:
                return tf.contrib.tpu.TPUEstimatorSpec(
                    self.mode, predictions=predictions
                )
            else:
                return tf.estimator.EstimatorSpec(self.mode, predictions=predictions)
        else:
            # true_tag_ids = self.labels
            true_tag_ids = self.tag2id(self.labels, "labels")

            # print(pred_strings)
            # print(self.labels)

            loss = self.loss_layer(logits, true_tag_ids, nwords, crf_params)

            metrics = self.compute_metrics(
                true_tag_ids, pred_ids, num_tags, indices, nwords
            )

            if self.mode == tf.estimator.ModeKeys.EVAL:
                if self.params["use_tpu"]:

                    def my_metric_fn(true_tag_ids, pred_ids, num_tags, indices, nwords):
                        return self.compute_metrics(
                            true_tag_ids, pred_ids, num_tags, indices, nwords
                        )

                    return tf.contrib.tpu.TPUEstimatorSpec(
                        self.mode,
                        loss=loss,
                        eval_metrics=(
                            my_metric_fn,
                            [true_tag_ids, pred_ids, num_tags, indices, nwords],
                        ),
                    )
                else:
                    # set up metrics reporter
                    # from ioflow.hooks.metrics_reporter import metrics_report_hook
                    # hook_class = metrics_report_hook(self.params)
                    # hook_metrics = {"eval_loss": loss}
                    # hook_metrics.update({'_'.join(['eval', k]): v[0] for k, v in metrics.items()})
                    # hook = hook_class(hook_metrics)

                    # return tf.estimator.EstimatorSpec(
                    #     self.mode, loss=loss, eval_metric_ops=metrics, evaluation_hooks=[hook])

                    return tf.estimator.EstimatorSpec(
                        self.mode, loss=loss, eval_metric_ops=metrics
                    )

            elif self.mode == tf.estimator.ModeKeys.TRAIN:
                train_op = tf.train.AdamOptimizer(
                    **self.params.get("optimizer_params", {})
                ).minimize(loss, global_step=tf.train.get_or_create_global_step())
                if self.params["use_tpu"]:
                    train_op = tf.contrib.tpu.CrossShardOptimizer(train_op)

                if self.params["use_tpu"]:
                    return tf.contrib.tpu.TPUEstimatorSpec(
                        self.mode, loss=loss, train_op=train_op
                    )
                else:
                    # set up metrics reporter
                    # from ioflow.hooks.metrics_reporter import metrics_report_hook
                    # hook_class = metrics_report_hook(self.params)
                    # hook = hook_class({"train_loss": loss})

                    # return tf.estimator.EstimatorSpec(
                    #     self.mode, loss=loss, train_op=train_op, training_hooks=[hook])

                    return tf.estimator.EstimatorSpec(
                        self.mode, loss=loss, train_op=train_op
                    )
